/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.util.Date;

import org.apache.commons.lang3.time.DateUtils;

/**
 * Represents an optionally open-ended date range.
 * The date range has a start and an end, either of which may be null to indicate
 * open-endedness.
 * If both are null, then the date range extends to negative and positive infinity.
 * 
 * <p> Ranges are always "consistent": if both from and to dates are defined, then {@code from <= to}.
 * 
 * <p> Provides the ability to access this range in a number of different ways,
 * including inclusive and exclusive interpretations.
 * @author Malcolm Lett
 */
public class DateRange {
	private Date after; // start timestamp (exclusive)
	private Date from; // start timestamp (inclusive)
	private Date to; // end timestamp (inclusive)
	private Date before; // end timestamp (exclusive)
	
	/**
	 * Creates a range with the specified limits.
	 * <p> {@code from} may be both null to indicate a <i>start</i> open ended range,
	 * and {@code to} may be both null to indicate an <i>end</i> open ended range.
	 * @param from instant at start of range (inclusive)
	 * @param to instant at end of range (inclusive)
	 */
	public static final DateRange fromTo(Date from, Date to) {
		return new DateRange(null, from, to, null);
	}
	
	/**
	 * Creates an infinite date range.
	 * ie: it accepts all dates.
	 */
	public DateRange() {
	}
	
	/**
	 * Creates a range with the specified limits.
	 * <p> {@code after/from} may be both null to indicate a <i>start</i> open ended range,
	 * and {@code to/before} may be both null to indicate an <i>end</i> open ended range.
	 * At most one of {@code after} and {@code from} may be supplied,
	 * and at most one of {@code to} and {@code before} may be supplied.
	 * @param after instant before start of range (exclusive)
	 * @param from instant at start of range (inclusive)
	 * @param to instant at end of range (inclusive)
	 * @param before instant after end of range (exclusive)
	 * @throws IllegalArgumentException if supply an invalid combination of parameters
	 */
	public DateRange(Date after, Date from, Date to, Date before) {
		if (after != null && from != null) {
			throw new IllegalArgumentException("Only one of 'after' and 'from' may be supplied");
		}
		if (to != null && before != null) {
			throw new IllegalArgumentException("Only one of 'before' and 'to' may be supplied");
		}
		
		this.after  = (after  != null) ? after  : (from   != null) ? instantBefore(from)   : null;
		this.from   = (from   != null) ? from   : (after  != null) ? instantAfter(after)   : null;
		this.to     = (to     != null) ? to     : (before != null) ? instantBefore(before) : null;
		this.before = (before != null) ? before : (to     != null) ? instantAfter(to)      : null;
	}

	/**
	 * Gets a string representation of the range.
	 * Uses the standard format of the dates contained. 
	 */
	@Override
	public String toString() {
		StringBuilder buf = new StringBuilder();
		if (from != null) {
			buf.append(from);
		}
		if (to != null) {
			if (buf.length() > 0) buf.append(" ");
			buf.append("to ");
			buf.append(to);
		}
		return buf.toString();
	}

	/**
	 * Gets the date-time instant at the start of the date range.
	 * The date range is <em>inclusive</em> of the date returned here.
	 * @return null if open-ended at the start
	 */
	public Date getStart() {
		return getFrom();
	}
	
	/**
	 * Gets the 'end' of the date range, which is the date-time instant
	 * immediately after the end of the date range.
	 * The date range is <em>exclusive</em> of the date returned here,
	 * as per the same interpretation of 'end' in String operations.
	 * @return null if open-ended at the end
	 */
	public Date getEnd() {
		return getBefore();
	}

	/**
	 * Gets the date-time instant at the start of the date range.
	 * The date range is <em>inclusive</em> of the date returned here.
	 * @return null if open-ended at the start
	 */
	public Date getFrom() {
		return from;
	}
	
	/**
	 * Gets the date-time instant at the very end of the date range.
	 * The date range is <em>inclusive</em> of the date returned here.
	 * @return null if open-ended at the end
	 */
	public Date getTo() {
		return to;
	}

	/**
	 * Gets the date-time instant immediately before the start of the date range.
	 * The date range is <em>exclusive</em> of the date returned here.
	 * @return null if open-ended at the start
	 */
	public Date getAfter() {
		return after;
	}

	/**
	 * Gets the date-time instant immediately after the end of the date range.
	 * The date range is <em>exclusive</em> of the date returned here.
	 * @return null if open-ended at the end
	 */
	public Date getBefore() {
		return before;
	}

	// support method
	protected static Date instantBefore(Date date) {
		if (date instanceof PartialDate) {
			return ((PartialDate) date).instantBefore();
		}
		else {
			return DateUtils.addMilliseconds(date, -1);
		}
	}

	// support method
	protected static Date instantAfter(Date date) {
		if (date instanceof PartialDate) {
			return ((PartialDate) date).instantAfter();
		}
		else {
			return DateUtils.addMilliseconds(date, +1);
		}
	}
}
