/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.util.Calendar;
import java.util.EnumSet;
import java.util.Set;

/**
 * Each of the letters supported within SimpleDateFormat patterns.
 * @author Malcolm
 */
enum DateTimeFieldType {
	ERA("G", Calendar.ERA),
	YEAR("y", Calendar.YEAR),
	MONTH_IN_YEAR("M", Calendar.MONTH),
	WEEK_IN_YEAR("w", Calendar.WEEK_OF_YEAR),
	WEEK_IN_MONTH("W", Calendar.WEEK_OF_MONTH),
	DAY_IN_YEAR("D", Calendar.DAY_OF_YEAR),
	DAY_IN_MONTH("d", Calendar.DAY_OF_MONTH),
	DAY_OF_WEEK_IN_MONTH("F", Calendar.DAY_OF_WEEK_IN_MONTH),
	DAY_OF_WEEK("E", Calendar.DAY_OF_WEEK),
	AM_PM("a", Calendar.AM_PM),
	HOUR_IN_DAY_0("H", Calendar.HOUR_OF_DAY), // TODO: how to distinguish these apart?
	HOUR_IN_DAY_1("k", Calendar.HOUR_OF_DAY),
	HOUR_IN_AMPM_0("K", Calendar.HOUR), // TODO: how to distinguish these apart?
	HOUR_IN_AMPM_1("h", Calendar.HOUR),
	MINUTE("m", Calendar.MINUTE),
	SECOND("s", Calendar.SECOND),
	MILLISECOND("S", Calendar.MILLISECOND),
	TIMEZONE_TEXT("z", Calendar.ZONE_OFFSET), // TODO: how to distinguish these apart?
	TIMEZONE("Z", Calendar.ZONE_OFFSET);
	
	private String code;
	private int calendarField;
	
	private DateTimeFieldType(String code, int calendarField) {
		this.code = code;
		this.calendarField = calendarField;
	}

	@Override
	public String toString() {
		return name()+"(\""+code()+"\")";
	}
	
	public String code() {
		return code;
	}
	
	public int calendarField() {
		return calendarField;
	}
	
	public static DateTimeFieldType valueOfCode(String code) {
		for (DateTimeFieldType letter: values()) {
			if (letter.code().equals(code)) {
				return letter;
			}
		}
		throw new IllegalArgumentException("Unexpected code "+code);
	}

	public static DateTimeFieldType valueOfCalendarField(int calendarField) {
		for (DateTimeFieldType letter: values()) {
			if (letter.calendarField() == calendarField) {
				return letter;
			}
		}
		throw new IllegalArgumentException("Unexpected calendar field "+calendarField);
	}
	
	public static String codeString() {
		return codeStringOf(values());
	}

	public static String codeStringOf(DateTimeFieldType... datePatternLetters) {
		StringBuilder buf = new StringBuilder();
		for (DateTimeFieldType letter: datePatternLetters) {
			buf.append(letter.code());
		}
		return buf.toString();
	}
	
	public static String codeStringOf(Set<DateTimeFieldType> datePatternLetters) {
		StringBuilder buf = new StringBuilder();
		for (DateTimeFieldType letter: datePatternLetters) {
			buf.append(letter.code());
		}
		return buf.toString();
	}
	
	/**
	 * Gets a set of just those codes that are related to the time,
	 * excluding time-zone.
	 */
	public static Set<DateTimeFieldType> timeValues() {
		return EnumSet.of(AM_PM, HOUR_IN_AMPM_0, HOUR_IN_AMPM_1, HOUR_IN_DAY_0, HOUR_IN_DAY_1, MINUTE, SECOND, MILLISECOND);
	}

	/**
	 * Gets a set of just those codes that are related to timezone.
	 */
	public static Set<DateTimeFieldType> timezoneValues() {
		return EnumSet.of(TIMEZONE, TIMEZONE_TEXT);
	}
}