/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.text.DateFormat;
import java.text.DateFormatSymbols;
import java.text.FieldPosition;
import java.text.Format;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.EnumSet;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;

import net.sourceforge.tedhi.DatePattern.Tag;

import org.apache.commons.lang3.time.DateUtils;
import org.apache.log4j.Logger;

/**
 * Date formatter and parser that recognises dates formatted according to
 * common human formatting, including recognition of partial dates where
 * one or more components are omitted.
 *
 * <p> Date formats are usually created via calls to the various {@code getDateInstance} class methods,
 * based on default or given locales and a number of formatting styles.
 * Styles supported are the FULL, LONG, MEDIUM, and SHORT styles defined within {@link DateFormat}. 
 * 
 * <p> For greater format control, date formats may be specified by <i>date pattern</i> strings.
 * These are the same as in {@link SimpleDateFormat}, with the exception that time-related fields
 * are not supported (am/pm, hour, minute, second, millisecond, or timezone).
 * 
 * <p> When parsing, date components may be separated by hyphens, slashes or spaces,
 * regardless of the specified or locale-defaulted date pattern.
 * Component separators may also be omitted, provided it parses to an unambiguous date.
 * Months may be numerals or names. Years may be two or four digits.
 * Any date component may be omitted, provided an unambiguous date can be derived.
 * Omitted components are defaulted based on the following rules:
 * <ul>
 * <li> Years are defaulted according to the {@link YearBias} (default {@link YearBias#NEAREST}
 * <li> Months are defaulted to the start or end of the year, depending on the {@link Bias} (default {@link Bias#START})
 * <li> Days are defaulted to the start or end of the month, depending on the {@link Bias} (default {@link Bias#START})
 * <li> Time-of-date is defaulted to the start of end of the day, depending on the {@link Bias} (default {@link Bias#START})
 * </ul>
 * 
 * <p>Parsing and formatting of time components are currently not supported.
 * An exception will be thrown by the {@code getInstance} methods if passed
 * a pattern containing time-related fields.
 * It is intended that support for time be added in a later version.
 * 
 * <p>Warning: instances of this class are not thread safe. Using the same instance
 * to parse date between multiple threads simultaneously will produce unpredictable results.
 * @author Malcolm Lett
 */
/* TODO: implement new method for handling multiple different kinds of separators.
 * 1) Use the locale's separator as the default internal separator (instead of "-").
 * 1a) If the locale doesn't have a separator between any two components then default to "-".
 * 2) Internally localize parserItem patterns of "d-m-y" etc. by
 *    starting with the original localized pattern and selecting out only the required components
 *    (ie: datePattern.withRetainedFields())
 * 2a) For each consecutive pair of tags with no separator, insert "-".
 * 3) For each date string being parsed, replace common separator characters with the one from (1/1a).
 * 3a) --> Don't know how to handle where separators are different between different components (eg: "h' h 'm' m 's' s'")
 * 3b) The first step is to generate one parserItem for each common separator, as well as the locale's separator.
 *     However that gets messy very quickly if want to handle where user has used different non-locale separators.
 *     
 * Note: this method will still work just as well as the current method for NZ style dates.
 * Except for the new approach potentially being less flexible with regard to use of mixed non-locale separators.
 */
public class FlexibleDateFormat extends Format {
	private static final long serialVersionUID = 1L;
	private static final Logger log = Logger.getLogger(FlexibleDateFormat.class);
	private static final YearBias DEFAULT_YEAR_BIAS = YearBias.NEAREST;
	private static final Bias DEFAULT_BIAS = Bias.START;
	private static final DefinedLevel BROADEST_DEFINITION = DefinedLevel.YEAR;
	private static final DefinedLevel FINEST_DEFINITION = DefinedLevel.DAY;
	static Date CURRENT_DATE_OVERRIDE = null; // used by unit tests only 

	private final DatePattern pattern;
	private final Locale locale;
	private final DateFormatSymbols formatSymbols;
	private TimeZone timeZone;
	private final YearBias yearBias; // year defaulting bias
	private final Bias bias; // day/month defaulting bias
	private boolean formatPartially = true;
	private final List<ParserItem> parsers;
	private final List<FormatterItem> formatters;
	private final FormatterItem fullFormatter;
	private final FormatterItem dayFormatter;
	private final FormatterItem monthFormatter;
	private final FormatterItem yearFormatter;

	/**
	 * Get a default date formatter that uses the default MEDIUM style for date.
	 * <p> Same as returned by {@link #getDateInstance()}.
	 * <p> <i>Note:</i> this method is subject to change. When time is supported
	 * this method will return a date/time formatter instead.
	 */
	public static FlexibleDateFormat getDefaultInstance() {
		return getDateInstance();
	}

	/**
	 * Gets the date formatter with the given pattern
	 * and the default date format symbols for the default locale.
	 * @param pattern the pattern describing the date and time format
	 * @throws NullPointerException if the given pattern is null
	 */
	public static FlexibleDateFormat getPatternInstance(String pattern) {
		return new FlexibleDateFormat(new DatePattern(pattern), Locale.getDefault());
	}

	/**
	 * Gets the date formatter with the given pattern
	 * and the default date format symbols for the given locale.
	 * @param pattern the pattern describing the date and time format
	 * @param locale the locale whose date format symbols should be used
	 * @throws NullPointerException if the given pattern or locale is null
	 */
	public static FlexibleDateFormat getPatternInstance(String pattern, Locale locale) {
		return new FlexibleDateFormat(new DatePattern(pattern), locale);
	}

	/**
	 * Gets the date formatter with the given pattern and date format symbols.
	 * @param pattern the pattern describing the date and time format
	 * @param formatSymbols the date format symbols to be used for formatting   
	 * @throws NullPointerException if the given pattern or formatSymbols is null
	 */
	public static FlexibleDateFormat getPatternInstance(String pattern, DateFormatSymbols formatSymbols) {
		return new FlexibleDateFormat(new DatePattern(pattern), Locale.getDefault(), formatSymbols);
	}
	
	/**
	 * Get a default date formatter that uses the default MEDIUM style for date.
	 * @return a date formatter.
	 */
	public static final FlexibleDateFormat getDateInstance() {
		return new FlexibleDateFormat(DatePattern.getDateInstance(), Locale.getDefault());
	}

	/**
	 * Gets the date formatter with the given formatting style for the default locale.
	 * @param style the given formatting style, defined within the {@link DateFormat} class.
	 *        For example, {@link DateFormat#SHORT} for "M/d/yy" in the US locale.
	 * @throws NullPointerException if the given style is null
	 * @return a date formatter.
	 */
	public static FlexibleDateFormat getDateInstance(int style) {
		return getDateInstance(style, Locale.getDefault());
	}

	/**
	 * Gets the date formatter with the given formatting style for the given locale.
	 * @param style the given formatting style, defined within the {@link DateFormat} class.
	 *        For example, {@link DateFormat#SHORT} for "M/d/yy" in the US locale.
	 * @param locale the given locale
	 * @throws NullPointerException if the given style or locale is null
	 * @return a date formatter.
	 */
	public static FlexibleDateFormat getDateInstance(int style, Locale locale) {
		return new FlexibleDateFormat(DatePattern.getDateInstance(style, locale), locale);
	}
	
	/**
	 * Returns an array of all locales for which the
	 * <code>get*Instance</code> methods of this class can return
	 * localized instances.
	 * The returned array represents the union of locales supported by the Java
	 * runtime and by installed
	 * {@link java.text.spi.DateFormatProvider DateFormatProvider} implementations.
	 * It must contain at least a <code>Locale</code> instance equal to
	 * {@link java.util.Locale#US Locale.US}.
	 *
	 * @return An array of locales for which localized
	 *         <code>FlexibleDateFormat</code> instances are available.
	 */
	public static Locale[] getAvailableLocales() {
		return DateFormat.getAvailableLocales();
	}
	
	// intermediate constructor
	protected FlexibleDateFormat(DatePattern pattern, Locale locale) {
		this(pattern, locale, new DateFormatSymbols(locale), DEFAULT_YEAR_BIAS, DEFAULT_BIAS);
	}
	
	// intermediate constructor
	protected FlexibleDateFormat(DatePattern pattern, Locale locale, DateFormatSymbols formatSymbols) {
		this(pattern, locale, formatSymbols, DEFAULT_YEAR_BIAS, DEFAULT_BIAS);
	}
	
	/**
	 * Constructs a {@code FlexibleDateFormat} using the given pattern
	 * and date format symbols, and the given bias when defaulting months and days etc.
	 * @param pattern the pattern describing the date and time format
	 * @param locale the locale whose date format symbols should be used
	 * @param yearBias how years should be defaulted   
	 * @param bias the month and day defaulting bias
	 * @param formatSymbols the date format symbols to be used for formatting   
	 * @throws NullPointerException if the given pattern, locale, formatSymbols, yearBias or bias is null
	 * @throws IllegalArgumentException if the given pattern is invalid 
	 */
	protected FlexibleDateFormat(DatePattern pattern, Locale locale, DateFormatSymbols formatSymbols,
			YearBias yearBias, Bias bias) {
		if (pattern == null) {
			throw new NullPointerException("pattern must be specified");
		}
		if (locale == null) {
			throw new NullPointerException("locale must be specified");
		}
		if (formatSymbols == null) {
			throw new NullPointerException("formatSymbols must be specified");
		}
		if (yearBias == null) {
			throw new NullPointerException("yearBias must be specified");
		}
		if (bias == null) {
			throw new NullPointerException("bias must be specified");
		}
		this.yearBias = yearBias;
		this.bias = bias;
		this.pattern = pattern;
		this.locale = locale;
		this.formatSymbols = formatSymbols;
		this.timeZone = TimeZone.getDefault(); // initial time-zone
		
		// Removed (see ParserItem.<init>):
		// According to comments in java.text.SimpleDateFormat:
		// The format object must be constructed using the symbols for this zone.
		// However, the calendar should use the current default TimeZone.
		// If this is not contained in the locale zone strings, then the zone
		// will be formatted using generic GMT+/-H:MM nomenclature.
		//this.calendar = Calendar.getInstance(TimeZone.getDefault(), locale);
		
		// Removed (see ParserItem.<init>):
		// Prepare number format in the same way that SimpleDateFormat uses.
		//this.numberFormat = NumberFormat.getIntegerInstance(locale);
		//this.numberFormat.setGroupingUsed(false);
		
		// parse pattern
		ParserItemFactory parser = new ParserItemFactory(pattern, formatSymbols, null, null);
		
		// prepare defaulting mechanisms
		DatePostProcessor defaultTime  = defaultInstantOfDay(bias);
		DatePostProcessor defaultDay   = defaultDayOfMonth(bias);
		DatePostProcessor defaultMonth = defaultMonthOfYear(bias);
		DatePostProcessor defaultYear  = defaultYear(yearBias);
		
		// construct parsing formats
		// full dates
		parsers = new ArrayList<ParserItem>();
		if (pattern.hasYear() && pattern.hasMonth() && pattern.hasDay()) {
			parsers.add(parser.ofPattern("d-M-yy").with(defaultTime));
			parsers.add(parser.ofPattern("d-MMM-yy").with(defaultTime));
			parsers.add(parser.ofPattern("ddMMyy").with(lengthExactly(6)).with(defaultTime));
			parsers.add(parser.ofPattern("ddMMyyyy").with(lengthExactly(8)).with(defaultTime));
		}

		// month and year only (defaults to biased day)
		if (pattern.hasYear() && pattern.hasMonth()) {
			parsers.add(parser.ofPattern("M-yy").with(yearSensible()).with(defaultDay, defaultTime).definedTo(DefinedLevel.MONTH));
			parsers.add(parser.ofPattern("MMM-yy").with(yearSensible()).with(defaultDay, defaultTime).definedTo(DefinedLevel.MONTH));
		}
		
		// day and month only (defaults to current year)
		if (pattern.hasMonth() && pattern.hasDay()) {
			parsers.add(parser.ofPattern("d-M").with(defaultYear, defaultTime).definedFrom(DefinedLevel.MONTH));
			parsers.add(parser.ofPattern("d-MMM").with(defaultYear, defaultTime).definedFrom(DefinedLevel.MONTH));
		}

		// day only (month biased, current year)
		// (only if both month and year excluded)
		if (pattern.hasDay()) {
			if (!pattern.hasMonth() && !pattern.hasYear()) {
				parsers.add(parser.ofPattern("d").with(defaultYear, defaultMonth, defaultTime).definedFrom(DefinedLevel.DAY).definedTo(DefinedLevel.DAY));
			}
		}
		
		// month only (day biased, current year)
		// (month names only to avoid confusion, unless year is excluded)
		if (pattern.hasMonth()) {
			if (!pattern.hasYear()) {
				parsers.add(parser.ofPattern("M").with(defaultYear, defaultDay, defaultTime).definedFrom(DefinedLevel.MONTH).definedTo(DefinedLevel.MONTH));
			}
			parsers.add(parser.ofPattern("MMM").with(defaultYear, defaultDay, defaultTime).definedFrom(DefinedLevel.MONTH).definedTo(DefinedLevel.MONTH));
		}
		
		// year only (day and month biased)
		if (pattern.hasYear()) {
			parsers.add(parser.ofPattern("yy").with(lengthExactly(2)).with(defaultMonth, defaultDay, defaultTime).definedTo(DefinedLevel.YEAR));
			parsers.add(parser.ofPattern("yyyy").with(lengthExactly(4)).with(defaultMonth, defaultDay, defaultTime).definedTo(DefinedLevel.YEAR));
		}
		
		// construct to-string formats
		FormatterItemFactory formatter = new FormatterItemFactory(pattern, formatSymbols, null, null);
		fullFormatter  = formatter.ofPreLocalizedPattern(pattern);
		dayFormatter   = fullFormatter;
		monthFormatter = formatter.ofPattern("M/y");
		yearFormatter  = formatter.ofPattern("y");
		formatters = new ArrayList<FormatterItem>(Arrays.asList(dayFormatter, monthFormatter, yearFormatter));
	}
	
	/**
	 * Gets a new instance that is a clone of this one plus it is configured
	 * with the supplied year bias.
	 * @param yearBias how years should be defaulted   
	 * @return the new instance
	 * @throws NullPointerException if the given yearBias is null
	 */
	public FlexibleDateFormat withYearBias(YearBias yearBias) {
		return new FlexibleDateFormat(pattern, locale, formatSymbols, yearBias, bias);
	}

	/**
	 * Gets a new instance that is a clone of this one plus it is configured
	 * with the supplied bias for months and days, etc.
	 * @param bias how months, days and time-of-day should be defaulted   
	 * @return the new instance
	 * @throws NullPointerException if the given yearBias is null
	 */
	public FlexibleDateFormat withBias(Bias bias) {
		return new FlexibleDateFormat(pattern, locale, formatSymbols, yearBias, bias);
	}
	
	/**
	 * Gets the pattern this date format is working to.
	 */
	@Override
	public String toString() {
		return fullFormatter.name();
	}

	/**
	 * Gets whether partial dates are formatted partially or fully.
	 */
	public boolean isFormatPartially() {
		return formatPartially;
	}

	/**
	 * Sets whether partial dates are formatted partially or fully.
	 * This applies to the least-significant components of the date only. 
	 * @param formatPartially
	 */
	public void setFormatPartially(boolean formatPartially) {
		this.formatPartially = formatPartially;
	}

	/**
	 * Indicates the bias being applied when defaulting years.
	 * @return the year bias
	 */
	public YearBias getYearBias() {
		return yearBias;
	}
	
	/**
	 * Indicates the bias being applied when defaulting days and months.
	 * @return the day and month bias
	 */
	public Bias getBias() {
		return bias;
	}
	
//Removed because not working, see comments in ParserItem.<init>
//	/**
//	 * Set the calendar to be used by this date format.  Initially, the default
//	 * calendar for the specified or default locale is used.
//	 *
//	 * <p>Any {@link java.util.TimeZone TimeZone} and
//	 * {@linkplain #isLenient() leniency} values that have previously been set are
//	 * overwritten by {@code newCalendar}'s values.
//	 * 
//	 * <p> Note: the instance provided is cloned internally,
//	 * so subsequent changes will not be reflected.
//	 *
//	 * @param newCalendar the new {@code Calendar} to be used by the date format
//	 */
//	// defensive programming: cloning in case multiple internal instances
//	// of SimpleDateFormat will conflict otherwise; but this could cause problems
//	// if external callers need the same instance to be used so they can manipulate
//	// the manipulatable parts of Calendar.
//	@Override
//	public void setCalendar(Calendar newCalendar) {
//		this.calendar = newCalendar;
//		for (FormatterItem formatter: formatters) {
//			formatter.formatter.setCalendar((Calendar)newCalendar.clone());
//		}
//		for (ParserItem parser: parsers) {
//			parser.parser.setCalendar((Calendar)newCalendar.clone());
//		}
//	}

//Removed because not working, see comments in ParserItem.<init>
//	/**
//	 * Allows you to set the number formatter.
//	 * 
//	 * <p> Note: the instance provided is cloned internally,
//	 * so subsequent changes will not be reflected.
//	 * @param newNumberFormat the given new NumberFormat.
//	 */
//	// defensive programming: cloning in case multiple internal instances
//	// of SimpleDateFormat will conflict otherwise; but this could cause problems
//	// if external callers need the same instance to be used so they can manipulate
//	// the manipulatable parts of NumberFormat.
//	@Override
//	public void setNumberFormat(NumberFormat newNumberFormat) {
//		this.numberFormat = newNumberFormat;
//		for (FormatterItem formatter: formatters) {
//			formatter.formatter.setNumberFormat((NumberFormat)newNumberFormat.clone());
//		}
//		for (ParserItem parser: parsers) {
//			parser.parser.setNumberFormat((NumberFormat)newNumberFormat.clone());
//		}
//	}

	/**
	 * Sets the time zone in use when formatting and parsing.
	 *
	 * <p>The {@code TimeZone} set by this method may be overwritten as
	 * a result of a call to the parse method.
	 * @param zone the given new time zone.
	 */
	// TODO: Interesting comment copied from DateFormat.setTimeZone(),
	//       do we need to reset the timezone on each ParserItem before parsing?
	public void setTimeZone(TimeZone zone) {
		this.timeZone = zone;
		for (FormatterItem formatter: formatters) {
			formatter.formatter.setTimeZone(zone);
		}
		for (ParserItem parser: parsers) {
			parser.parser.setTimeZone(zone);
		}
	}
	
	/**
	 * Gets the time zone.
	 * @return the time zone in use when formatting and parsing.
	 */
	public TimeZone getTimeZone() {
		return timeZone;
	}

	/**
	 * Overrides hashCode.
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((bias == null) ? 0 : bias.hashCode());
		result = prime * result + (formatPartially ? 1231 : 1237);
		result = prime * result
				+ ((formatSymbols == null) ? 0 : formatSymbols.hashCode());
		result = prime * result + ((locale == null) ? 0 : locale.hashCode());
		result = prime * result + ((pattern == null) ? 0 : pattern.hashCode());
		result = prime * result
				+ ((timeZone == null) ? 0 : timeZone.hashCode());
		result = prime * result
				+ ((yearBias == null) ? 0 : yearBias.hashCode());
		return result;
	}

	/**
	 * Overrides equals.
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof FlexibleDateFormat)) {
			return false;
		}
		FlexibleDateFormat other = (FlexibleDateFormat) obj;
		if (bias != other.bias) {
			return false;
		}
		if (formatPartially != other.formatPartially) {
			return false;
		}
		if (formatSymbols == null) {
			if (other.formatSymbols != null) {
				return false;
			}
		} else if (!formatSymbols.equals(other.formatSymbols)) {
			return false;
		}
		if (locale == null) {
			if (other.locale != null) {
				return false;
			}
		} else if (!locale.equals(other.locale)) {
			return false;
		}
		if (pattern == null) {
			if (other.pattern != null) {
				return false;
			}
		} else if (!pattern.equals(other.pattern)) {
			return false;
		}
		if (timeZone == null) {
			if (other.timeZone != null) {
				return false;
			}
		} else if (!timeZone.equals(other.timeZone)) {
			return false;
		}
		if (yearBias != other.yearBias) {
			return false;
		}
		return true;
	}
	
	/**
	 * Overrides Cloneable.
	 * Not a complete clone implementation.
	 * Probably best not to use.
	 */
	@Override
	public FlexibleDateFormat clone() {
		FlexibleDateFormat other = (FlexibleDateFormat) super.clone();
		return other;
		
		// doesn't work because all these fields are declared 'final' 
//		FlexibleDateFormat other = (FlexibleDateFormat) super.clone();
//		other.bias = this.bias;
//		other.dayFormatter = this.dayFormatter;
//		other.formatPartially = this.formatPartially;
//		other.formatSymbols = this.formatSymbols;
//		other.formatters = this.formatters;
//		other.fullFormatter = this.fullFormatter;
//		other.locale = this.locale;
//		other.monthFormatter = this.monthFormatter;
//		other.parsers = this.parsers;
//		other.pattern = this.pattern;
//		other.yearBias = this.yearBias;
//		other.yearFormatter = this.yearFormatter;
//		return other;
	}
	
	/**
	 * Formats a Date into a date/time string.
	 * @param date the time value to be formatted into a time string.
	 * @return the formatted time string.
	 */
	public String format(Date date) {
		return format(date, new StringBuffer(), new FieldPosition(DateFormat.YEAR_FIELD)).toString();
	}
	
	/**
	 * Overrides Format.
	 * Formats a time object into a time string. Examples of time objects
	 * are a time value expressed in milliseconds and a Date object.
	 * @param obj must be a Number or a Date.
	 * @param toAppendTo the string buffer for the returning time string.
	 * @return the string buffer passed in as toAppendTo, with formatted text appended.
	 * @param fieldPosition keeps track of the position of the field
	 * within the returned string.
	 * On input: an alignment field,
	 * if desired. On output: the offsets of the alignment field. For
	 * example, given a time text "1996.07.10 AD at 15:08:56 PDT",
	 * if the given fieldPosition is DateFormat.YEAR_FIELD, the
	 * begin index and end index of fieldPosition will be set to
	 * 0 and 4, respectively.
	 * Notice that if the same time field appears
	 * more than once in a pattern, the fieldPosition will be set for the first
	 * occurrence of that time field. For instance, formatting a Date to
	 * the time string "1 PM PDT (Pacific Daylight Time)" using the pattern
	 * "h a z (zzzz)" and the alignment field DateFormat.TIMEZONE_FIELD,
	 * the begin index and end index of fieldPosition will be set to
	 * 5 and 8, respectively, for the first occurrence of the timezone
	 * pattern character 'z'.
	 * @see java.text.Format
	 */
	@Override
	public StringBuffer format(Object obj, StringBuffer toAppendTo, FieldPosition fieldPosition) {
		if (obj instanceof Date) {
			return format( (Date)obj, toAppendTo, fieldPosition );
		}
		else if (obj instanceof Number) {
			return format( new Date(((Number)obj).longValue()), toAppendTo, fieldPosition );
		}
		else {
			throw new IllegalArgumentException("Cannot format given Object as a Date");
		}
	}
	
	/**
	 * Formats the date according to the level at which it is defined.
	 * @param date a Date to be formatted into a date/time string.
	 * @param toAppendTo the string buffer for the returning date/time string.
	 * @param fieldPosition keeps track of the position of the field
	 * within the returned string.
	 * On input: an alignment field, if desired.
	 * On output: the offsets of the alignment field. For
	 * example, given a time text "1996.07.10 AD at 15:08:56 PDT",
	 * if the given fieldPosition is DateFormat.YEAR_FIELD, the
	 * begin index and end index of fieldPosition will be set to
	 * 0 and 4, respectively.
	 * Notice that if the same time field appears
	 * more than once in a pattern, the fieldPosition will be set for the first
	 * occurrence of that time field. For instance, formatting a Date to
	 * the time string "1 PM PDT (Pacific Daylight Time)" using the pattern
	 * "h a z (zzzz)" and the alignment field DateFormat.TIMEZONE_FIELD,
	 * the begin index and end index of fieldPosition will be set to
	 * 5 and 8, respectively, for the first occurrence of the timezone
	 * pattern character 'z'.
	 * @return the string buffer passed in as toAppendTo, with formatted text appended.
	 */
	public StringBuffer format(Date date, StringBuffer toAppendTo, FieldPosition fieldPosition) {
		if (date == null) {
			toAppendTo.append("null");
			return toAppendTo;
		}
		
		DefinedLevel definedTo = FINEST_DEFINITION;
		if (date instanceof PartialDate) {
			definedTo = ((PartialDate) date).definedTo();
		}
		if (!formatPartially) {
			definedTo = FINEST_DEFINITION;
		}
		
		switch (definedTo) {
			case DAY:   toAppendTo.append(dayFormatter.format(date)); break;
			case MONTH: toAppendTo.append(monthFormatter.format(date)); break;
			case YEAR:  toAppendTo.append(yearFormatter.format(date)); break;
			default: throw new UnsupportedOperationException("Unexpected defined level ["+definedTo+"]");
		}
		return toAppendTo;
	}

	/**
	 * Formats an Object producing an <code>AttributedCharacterIterator</code>.
	 * You can use the returned <code>AttributedCharacterIterator</code>
	 * to build the resulting String, as well as to determine information
	 * about the resulting String.
	 * <p>
	 * Each attribute key of the AttributedCharacterIterator will be of type
	 * <code>DateFormat.Field</code>, with the corresponding attribute value
	 * being the same as the attribute key.
	 *
	 * @exception NullPointerException if obj is null.
	 * @exception IllegalArgumentException if the Format cannot format the
	 *            given object, or if the Format's pattern string is invalid.
	 * @param obj The object to format
	 * @return AttributedCharacterIterator describing the formatted value.
	 */
	@Override
	public AttributedCharacterIterator formatToCharacterIterator(Object obj) {
		if (obj == null) {
			AttributedString as = new AttributedString("null");
			return as.getIterator();
		}
		
		DefinedLevel definedTo = FINEST_DEFINITION;
		if (obj instanceof PartialDate) {
			definedTo = ((PartialDate) obj).definedTo();
		}
		if (!formatPartially) {
			definedTo = FINEST_DEFINITION;
		}
		
		switch (definedTo) {
			case DAY:   return dayFormatter.formatToCharacterIterator(obj);
			case MONTH: return monthFormatter.formatToCharacterIterator(obj);
			case YEAR:  return yearFormatter.formatToCharacterIterator(obj);
			default: throw new UnsupportedOperationException("Unexpected defined level ["+definedTo+"]");
		}
	}
	
	/**
	 * Parses the supplied text into a single date.
	 * @param source the string representation of the date to be parsed
	 * @return the parsed date object
	 * @throws ParseException on any parse errors
	 */
	public PartialDate parse(String source) throws ParseException {
		return parseInternal(source, new ParsePosition(0));
	}
	
	/**
	 * Parses the supplied text into a single date.
	 * @param source the string representation of the date to be parsed
	 * @param pos   On input, the position at which to start parsing; on
	 *              output, the position at which parsing terminated, or the
	 *              start position if the parse failed.
	 * @return A {@code PartialDate}, or {@code null} if the input could not be parsed
	 */
	public PartialDate parse(String source, ParsePosition pos) {
		try {
			return parseInternal(source, pos);
		} catch (ParseException dropped) {
			return null;
		}
	}
	
	/**
	 * Parses text from the beginning of the given string to produce an object.
	 * The method may not use the entire text of the given string.
	 *
	 * @param source A <code>String</code> whose beginning should be parsed.
	 * @return An <code>Object</code> parsed from the string.
	 * @exception ParseException if the beginning of the specified string
	 *            cannot be parsed.
	 */
	@Override
	public Object parseObject(String source) throws ParseException {
		return parseInternal(source, new ParsePosition(0));
	}
	
	/**
	 * Parses text from a string to produce a <code>Date</code>.
	 * <p>
	 * The method attempts to parse text starting at the index given by
	 * <code>pos</code>.
	 * If parsing succeeds, then the index of <code>pos</code> is updated
	 * to the index after the last character used (parsing does not necessarily
	 * use all characters up to the end of the string), and the parsed
	 * date is returned. The updated <code>pos</code> can be used to
	 * indicate the starting point for the next call to this method.
	 * If an error occurs, then the index of <code>pos</code> is not
	 * changed, the error index of <code>pos</code> is set to the index of
	 * the character where the error occurred, and null is returned.
	 * <p>
	 * See the {@link #parse(String, ParsePosition)} method for more information
	 * on date parsing.
	 *
	 * @param source A <code>String</code>, part of which should be parsed.
	 * @param pos A <code>ParsePosition</code> object with index and error
	 *            index information as described above.
	 * @return A <code>Date</code> parsed from the string. In case of
	 *         error, returns null.
	 * @exception NullPointerException if <code>pos</code> is null.
	 */
	@Override
	public Object parseObject(String source, ParsePosition pos) {
		try {
			return parseInternal(source, pos);
		} catch (ParseException dropped) {
			return null;
		}
	}
	
	/**
	 * Parses the supplied text into a single date.
	 * @param text the string representation of the date to be parsed
	 * @param pos   On input, the position at which to start parsing; on
	 *              output or exception, the position at which parsing terminated, or the
	 *              start position if the parse failed.
	 * @return the date if successfully parsed
	 * @throws ParseException on any parse errors
	 */
	protected PartialDate parseInternal(String text, ParsePosition pos) throws ParseException {
		int posIndexOffset = pos.getIndex();
		String origPattern = text.substring(posIndexOffset);
		text = origPattern.trim();
		text = text.replace("/", "-");
		text = text.replace("\\", "-");
		text = text.replaceAll("\\p{Space}+", "-");
		
		List<ParserItem> resultParsers = new ArrayList<ParserItem>();
		List<PartialDate> results = new ArrayList<PartialDate>();
		List<ParsePosition> resultPoss = new ArrayList<ParsePosition>();
		for (ParserItem parser: parsers) {
			try {
				ParsePosition tmpPos = new ParsePosition(pos.getIndex());
				Date date = parser.parse(text, tmpPos);
				if (date == null) {
					throw new ParseException("date was null", posIndexOffset);
				}
				
				if (log.isDebugEnabled()) {
					if (text.equals(origPattern)) {
						log.debug("Parsed ["+origPattern+"] with format ["+parser.name()+"] as "+date);
					}
					else {
						log.debug("Parsed ["+origPattern+"] as ["+text+"] with format ["+parser.name()+"] as "+date);
					}
				}
				results.add(new PartialDate(date, parser.definedFromLevel(), parser.definedToLevel()));
				resultPoss.add(tmpPos);
				resultParsers.add(parser);
			} catch (ParseException dropped) {
				// skip to next one
			}
		}
		if (results.size() > 1) {
			pos.setErrorIndex(posIndexOffset);
			throw new ParseException("Ambiguous date '"+text+"' recognised as formats "+resultParsers,posIndexOffset);
		}
		if (!results.isEmpty()) {
			ParsePosition resultPos = resultPoss.get(0);
			pos.setIndex(resultPos.getIndex()+posIndexOffset);
			pos.setErrorIndex(resultPos.getErrorIndex()+posIndexOffset);
			return results.get(0);
		}
		
		// no formats matched
		pos.setErrorIndex(posIndexOffset);
		throw new ParseException("Unrecognised date '"+origPattern+"'", posIndexOffset);
	}

	/**
	 * Internal support method.
	 */
	// TODO: consider whether need to apply timezone smarts as in java.text.SimpleDateFormat.initializeCalendar(Locale)
	private static Calendar calendarOf(Date date) {
		Calendar calendar = Calendar.getInstance();
		calendar.setTime(date);
		return calendar;
	}
	
	/**
	 * Internal support method to get current date today,
	 * with time component truncated.
	 * Overridden by unit tests where needed.
	 */
	// note: CURRENT_DATE_OVERRIDE access is not done in a thread-safe way
	// but it doesn't need to be because it's only used in single-threaded unit tests
	protected static Date today() {
		Date date;
		if (CURRENT_DATE_OVERRIDE != null) {
			date = CURRENT_DATE_OVERRIDE;
		}
		else {
			date = new Date();
		}
		return DateUtils.truncate(date, Calendar.DAY_OF_MONTH);
	}
	
	/**
	 * Makes it easier to write the code that constructs the ParserItems.
	 * In particular, this actually translates patterns for according to
	 * configured localization prior to passing them to the ParserItem
	 * constructors.
	 */
	static class ParserItemFactory {
		private DatePattern localizedPattern; // user requested pattern
		private DateFormatSymbols formatSymbols;
		private Calendar calendar; // may be null
		private NumberFormat numberFormat; // may be null
		
		public ParserItemFactory(DatePattern localizedPattern, DateFormatSymbols formatSymbols,
				Calendar calendar, NumberFormat numberFormat) {
			this.localizedPattern = localizedPattern;
			this.formatSymbols = formatSymbols;
			this.calendar = calendar;
			this.numberFormat = numberFormat;
		}
		
		public ParserItem ofPattern(String unlocalizedPattern) {
			return ParserItem.ofPattern(localizedOrderOf(unlocalizedPattern), formatSymbols,
					(calendar == null) ? null : (Calendar)calendar.clone(),
					(numberFormat == null) ? null : (NumberFormat)numberFormat.clone());
		}

		DatePattern localizedOrderOf(String unlocalizedPattern) {
			return localizedOrderOf(new DatePattern(unlocalizedPattern));
		}
		
		/**
		 * Takes the simple unlocalized proposed pattern and re-arranges its components
		 * to meet those of the configured locale.
		 * Only the order is changed, the separators are left as-is.
		 * @param unlocalizedPattern proposed pattern to be re-arranged
		 * @return updated pattern
		 */
		DatePattern localizedOrderOf(DatePattern unlocalizedPattern) {
			// TODO: should check for multiple instances of a given type
			List<DateTimeFieldType> fieldOrder = new ArrayList<DateTimeFieldType>();
			for (Tag tag: localizedPattern.tags()) {
				fieldOrder.add(tag.type());
			}
			
			return unlocalizedPattern.withFieldsOrderedAs(fieldOrder);
		}
	}

	/**
	 * Makes it easier to write the code that constructs the FormatterItems.
	 * In particular, this actually translates patterns for according to
	 * configured localization prior to passing them to the FormatterItem
	 * constructors.
	 */
	static class FormatterItemFactory {
		private DatePattern localizedPattern; // user requested pattern
		private DateFormatSymbols formatSymbols;
		private Calendar calendar; // may be null
		private NumberFormat numberFormat; // may be null
		
		public FormatterItemFactory(DatePattern localizedPattern, DateFormatSymbols formatSymbols,
				Calendar calendar, NumberFormat numberFormat) {
			this.localizedPattern = localizedPattern;
			this.formatSymbols = formatSymbols;
			this.calendar = calendar;
			this.numberFormat = numberFormat;
		}
		
		public FormatterItem ofPattern(String unlocalizedPattern) {
			return FormatterItem.ofPattern(localizedPatternGiven(unlocalizedPattern), formatSymbols,
					(calendar == null) ? null : (Calendar)calendar.clone(),
					(numberFormat == null) ? null : (NumberFormat)numberFormat.clone());
		}

		public FormatterItem ofPreLocalizedPattern(DatePattern localizedPattern) {
			return FormatterItem.ofPattern(localizedPattern, formatSymbols,
					(calendar == null) ? null : (Calendar)calendar.clone(),
					(numberFormat == null) ? null : (NumberFormat)numberFormat.clone());
		}

		public FormatterItem of(DateFormat dateFormat) {
			return FormatterItem.of(dateFormat);
		}

		DatePattern localizedPatternGiven(String unlocalizedPattern) {
			return localizedPatternGiven(new DatePattern(unlocalizedPattern));
		}
		
		/**
		 * Takes the simple unlocalized proposed pattern and uses it to select
		 * the components from the configured locale.
		 * The order and size of components and their separators are left as per the locale.
		 * @param unlocalizedPattern pattern for defining component 
		 * @return the generated pattern
		 */
		DatePattern localizedPatternGiven(DatePattern unlocalizedPattern) {
			// identify components to include
			EnumSet<DateTimeFieldType> retainedFieldTypes = EnumSet.allOf(DateTimeFieldType.class);
			//datePatternLetters.remove(DatePatternLetter.timeValues());
			//datePatternLetters.remove(DatePatternLetter.timezoneValues());
			retainedFieldTypes.remove(DateTimeFieldType.YEAR);
			retainedFieldTypes.remove(DateTimeFieldType.MONTH_IN_YEAR);
			retainedFieldTypes.remove(DateTimeFieldType.DAY_IN_MONTH);
			
			for (Tag includedTag: unlocalizedPattern.tags()) {
				retainedFieldTypes.add(includedTag.type());
			}
			
			DatePattern impliedDatePattern = localizedPattern.withRetainedFields(retainedFieldTypes);
			return impliedDatePattern;
		}
	}
	
	/**
	 * A single specific recognised partial or complete format used
	 * for parsing only.
	 * Used internally to wrap a SimpleDateFormat with extra metadata.
	 * Instances are mutable during construction of the parent object but
	 * should never be modified after that.
	 */
	protected static class ParserItem {
		private String name; // the pattern itself
		private DefinedLevel definedFromLevel = BROADEST_DEFINITION;
		private DefinedLevel definedToLevel = FINEST_DEFINITION;
		private DateFormat parser;
		private List<TextPreValidation> preValidations = new ArrayList<TextPreValidation>();
		private List<DatePostProcessor> postProcessors = new ArrayList<DatePostProcessor>();
		
		public static ParserItem ofPattern(DatePattern pattern, DateFormatSymbols formatSymbols,
				Calendar calendar, NumberFormat numberFormat) {
			ParserItem item = new ParserItem();
			item.name = pattern.toPatternString();
			item.parser = new SimpleDateFormat(pattern.toPatternString(), formatSymbols);
			item.parser.setLenient(false); // must always be false, or multiple parsers get hits
			
			// nice idea but it breaks SimpleDateFormat for some reason.
			// ie: "8-7-2010" gets parsed via "ddMMyyyy" pattern, when it shouldn't
			// Update: Turns out this is caused by the calendar having isLenient=true,
			// which breaks the assumptions made by this class.
			//if (calendar != null) {
			//	item.parser.setCalendar(calendar);
			//}
			//if (numberFormat != null) {
			//	item.parser.setNumberFormat(numberFormat);
			//}
			return item;
		}
		
		private ParserItem() { } // internal use only
		
		/**
		 * Adds the specified pre-validations
		 * @param additionalPreValidations
		 * @return this instance for method chaining
		 */
		public ParserItem with(TextPreValidation... additionalPreValidations) {
			for (TextPreValidation validation: additionalPreValidations) {
				preValidations.add(validation);
			}
			return this;
		}
		
		/**
		 * Adds the specified post processors
		 * @param additionalPostProcessors
		 * @return this instance for method chaining
		 */
		public ParserItem with(DatePostProcessor... additionalPostProcessors) {
			for (DatePostProcessor processor: additionalPostProcessors) {
				postProcessors.add(processor);
			}
			return this;
		}

		/**
		 * Adds the specified date defined level.
		 * @param fromLevel
		 * @return this instance for method chaining
		 */
		public ParserItem definedFrom(DefinedLevel fromLevel) {
			this.definedFromLevel = fromLevel;
			return this;
		}
		
		/**
		 * Adds the specified date defined level.
		 * @param toLevel
		 * @return this instance for method chaining
		 */
		public ParserItem definedTo(DefinedLevel toLevel) {
			this.definedToLevel = toLevel;
			return this;
		}
		
		@Override
		public String toString() {
			return name();
		}
		
		public String name() {
			return name;
		}

		public DefinedLevel definedFromLevel() {
			return definedFromLevel;
		}
		
		public DefinedLevel definedToLevel() {
			return definedToLevel;
		}

		/**
		 * Parses the specified text if possible.
		 * @param source
		 * @throws ParseException on any parse error
		 */
		public Date parse(String source, ParsePosition pos) throws ParseException {
			for (TextPreValidation validation: preValidations) {
				if (!validation.accepts(source)) {
					throw new ParseException("doesn't meet pre-validations",0);
				}
			}

			// attempt parse
			Date date = parser.parse(source, pos);
			if (pos.getIndex() < source.length()) {
				// DateFormat accepts this as OK by default
				throw new ParseException("didn't consume end of source", pos.getIndex());
			}
			
			// validate result
			for (DatePostProcessor postProcessor: postProcessors) {
				date = postProcessor.process(date);
			}
			return date;
		}
	}

	/**
	 * A single specific partial or complete format used for formatting only.
	 * Used internally to wrap a SimpleDateFormat with extra metadata.
	 * Instances are mutable during construction of the parent object but
	 * should never be modified after that.
	 */
	private static class FormatterItem {
		private String name; // the pattern itself
		private DateFormat formatter;

		public static FormatterItem ofPattern(DatePattern pattern, DateFormatSymbols formatSymbols,
				Calendar calendar, NumberFormat numberFormat) {
			FormatterItem item = new FormatterItem();
			item.name = pattern.toPatternString();
			item.formatter = new SimpleDateFormat(pattern.toPatternString(), formatSymbols);
			
			// nice idea but it didn't work for ParserItem, so removing it here too
			//if (calendar != null) {
			//	item.formatter.setCalendar(calendar);
			//}
			//if (numberFormat != null) {
			//	item.formatter.setNumberFormat(numberFormat);
			//}
			return item;
		}

		public static FormatterItem of(DateFormat dateFormat) {
			FormatterItem item = new FormatterItem();
			item.name = dateFormat.toString();
			item.formatter = dateFormat;
			return item;
		}
		
		private FormatterItem() { } // internal use only

		@Override
		public String toString() {
			return name();
		}
		
		public String name() {
			return name;
		}
		
		public String format(Date date) {
			return formatter.format(date);
		}
		
		public AttributedCharacterIterator formatToCharacterIterator(Object obj) {
			return formatter.formatToCharacterIterator(obj);
		}
	}
	
	/**
	 * Gets a post processor that defaults the year
	 * according to the given bias.
	 */
	protected static DatePostProcessor defaultYear(YearBias yearBias) {
		switch (yearBias) {
			case CURRENT: return currentYear();
			case FUTURE:  return futureOrTodayYear();
			case PAST:    return pastOrTodayYear();
			case NEAREST: return nearestYear(); 
			default: throw new UnsupportedOperationException("Unexpected year bias "+yearBias);
		}
	}

	/**
	 * Gets a post processor that defaults the month-of-year according to the given bias.
	 */
	protected static DatePostProcessor defaultMonthOfYear(Bias bias) {
		switch (bias) {
			case START: return firstMonthOfYear();
			case END:   return lastMonthOfYear();
			default: throw new UnsupportedOperationException("Unexpected bias "+bias);
		}
	}
	
	/**
	 * Gets a post processor that defaults the day-of-month according to the given bias.
	 */
	protected static DatePostProcessor defaultDayOfMonth(Bias bias) {
		switch (bias) {
			case START: return firstDayOfMonth();
			case END:   return lastDayOfMonth();
			default: throw new UnsupportedOperationException("Unexpected bias "+bias);
		}
	}

	/**
	 * Gets a post processor that defaults the time-of-date according to the given bias.
	 */
	protected static DatePostProcessor defaultInstantOfDay(Bias bias) {
		switch (bias) {
			case START: return firstInstantOfDay();
			case END:   return lastInstantOfDay();
			default: throw new UnsupportedOperationException("Unexpected bias "+bias);
		}
	}
	
	/**
	 * Gets a post processor that sets the year to the current year.
	 */
	protected static DatePostProcessor currentYear() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar now = calendarOf(today());
				
				Calendar c = calendarOf(date);
				c.set(Calendar.YEAR, now.get(Calendar.YEAR));
				return c.getTime();
			}
		};
	}

	/**
	 * Gets a post processor that sets the year such that
	 * the date is either equal to or after the current day,
	 * but is as close as possible to the current day.
	 */
	protected static DatePostProcessor futureOrTodayYear() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar today = calendarOf(today());
				
				Calendar c = calendarOf(date);
				c.set(Calendar.YEAR, today.get(Calendar.YEAR));
				
				// fix if in past
				if (DateUtils.truncatedCompareTo(c, today, Calendar.DAY_OF_MONTH) < 0) {
					c.add(Calendar.YEAR, +1);
				}
				return c.getTime();
			}
		};
	}

	/**
	 * Gets a post processor that sets the year such that
	 * the date is either equal to or before the current day,
	 * but is as close as possible to the current day.
	 */
	protected static DatePostProcessor pastOrTodayYear() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar today = calendarOf(today());
				
				Calendar c = calendarOf(date);
				c.set(Calendar.YEAR, today.get(Calendar.YEAR));
				
				// fix if in future
				if (DateUtils.truncatedCompareTo(c, today, Calendar.DAY_OF_MONTH) > 0) {
					c.add(Calendar.YEAR, -1);
				}
				return c.getTime();
			}
		};
	}
	
	/**
	 * Gets a post processor that sets the year such
	 * that the date is nearest to the current instant in time. 
	 */
	protected static DatePostProcessor nearestYear() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar now = calendarOf(today());
				
				// get given current year
				Calendar c1 = calendarOf(date);
				c1.set(Calendar.YEAR, now.get(Calendar.YEAR));
				
				// now get "other" year (closest either after or before current)
				Calendar c2 = (Calendar)c1.clone();
				if (c2.compareTo(now) < 0) {
					c2.add(Calendar.YEAR, +1);
				}
				else {
					c2.add(Calendar.YEAR, -1);
				}
				
				// pick closest to now
				long c1Diff = Math.abs(now.getTimeInMillis() - c1.getTimeInMillis());
				long c2Diff = Math.abs(now.getTimeInMillis() - c2.getTimeInMillis());
				if (c1Diff <= c2Diff) {
					return c1.getTime();
				}
				else {
					return c2.getTime();
				}
			}
		};
	}
	
	/**
	 * Gets a post processor that sets the time to the start of the day
	 * of the supplied date.
	 */
	protected static DatePostProcessor firstInstantOfDay() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				// note: this implementation seems much simpler and less chance of
				// bugs than if using DateUtils.truncate()
				Calendar c = calendarOf(date);
				c.set(Calendar.HOUR_OF_DAY, 0);
				c.set(Calendar.MINUTE, 0);
				c.set(Calendar.SECOND, 0);
				c.set(Calendar.MILLISECOND, 0);
				return c.getTime();
			}
		};
	}
	
	/**
	 * Gets a post processor that sets the time to the end of the day
	 * of the supplied date.
	 * Is leap-second aware.
	 */
	protected static DatePostProcessor lastInstantOfDay() {
		// works by grabbing the timestamp on the 1st instant of the following day
		// and then deducting 1 instant.
		return new DatePostProcessor() {
			private DatePostProcessor firstInstantProcessor = firstInstantOfDay();
			
			@Override
			public Date process(Date date) throws ParseException {
				// get 1st instant of next day
				Date startOfDay = firstInstantProcessor.process(date);
				Date startOfNextDay = DateUtils.addDays(startOfDay, 1);
				
				// get end of this day
				Date endOfDay = DateUtils.addMilliseconds(startOfNextDay, -1);
				return endOfDay;
			}
		};
	}
	
	/**
	 * Gets a post processor that sets the day to the start of the month
	 * of the supplied date.
	 */
	protected static DatePostProcessor firstDayOfMonth() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar c = calendarOf(date);
				c.set(Calendar.DAY_OF_MONTH, 1);
				return c.getTime();
			}
		};
	}

	/**
	 * Gets a post processor that sets the day to the end of the month
	 * of the supplied date.
	 * Is leap-year aware.
	 */
	protected static DatePostProcessor lastDayOfMonth() {
		// works by grabbing the date on the 1st of the following month
		// and then deducting 1 day.
		return new DatePostProcessor() {
			private DatePostProcessor firstOfMonthProcessor = firstDayOfMonth();
			
			@Override
			public Date process(Date date) throws ParseException {
				// get 1st of next month
				Date startOfMonth = firstOfMonthProcessor.process(date);
				Date startOfNextMonth = DateUtils.addMonths(startOfMonth, 1);
				
				// get end of this month
				Date endOfMonth = DateUtils.addDays(startOfNextMonth, -1);
				return endOfMonth;
			}
		};
	}

	/**
	 * Gets a post processor that sets the month to January.
	 */
	protected static DatePostProcessor firstMonthOfYear() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar c = calendarOf(date);
				c.set(Calendar.MONTH, Calendar.JANUARY);
				return c.getTime();
			}
		};
	}

	/**
	 * Gets a post processor that sets the month to December.
	 */
	protected static DatePostProcessor lastMonthOfYear() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				Calendar c = calendarOf(date);
				c.set(Calendar.MONTH, Calendar.DECEMBER);
				return c.getTime();
			}
		};
	}

	/**
	 * Gets a post processor that validates that the year
	 * is sensible.
	 * If the year is parsed as a 2-digit number then it was probably
	 * meant as a month.
	 */
	protected static DatePostProcessor yearSensible() {
		return new DatePostProcessor() {
			@Override
			public Date process(Date date) throws ParseException {
				if (calendarOf(date).get(Calendar.YEAR) < 100) {
					throw new ParseException("wrong format, "+calendarOf(date).get(Calendar.YEAR)+" must have been a month",0);
				}
				return date;
			}
		};
	}
	
	/**
	 * Gets a pre-validation that checks the length of the source text.
	 */
	protected static TextPreValidation lengthExactly(final int length) {
		return new TextPreValidation() {
			@Override
			public boolean accepts(String text) throws ParseException {
				return (text != null) && text.length() == length;
			}
		};
	}

	/**
	 * Used to apply assertions about the text to explicitly prohibit the
	 * format from attempting to parse in cases where it would othewise produce
	 * an incorrect result.
	 */
	protected static interface TextPreValidation {
		/**
		 * Indicates whether the text is accepted by this validation rule.
		 * @param text the text that is about to be parsed.
		 * @return {@code false} if not accepted
		 * @throws ParseException on rare occasions when needed
		 */
		public boolean accepts(String text) throws ParseException;
	}
	
	/**
	 * Used to default portions of dates when they were specified
	 * as a partial date but need to be stored as a concrete
	 * full date.
	 * Can also apply post-validations if needed.
	 */
	protected static interface DatePostProcessor {
		public Date process(Date date) throws ParseException;
	}
}
