/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.text.DateFormat;
import java.text.DateFormatSymbols;
import java.text.FieldPosition;
import java.text.Format;
import java.text.ParseException;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sourceforge.tedhi.FlexibleDateFormat.DatePostProcessor;

import org.apache.log4j.Logger;

/**
 * Parser and formatter that recognises simple date ranges formatted according to common human formatting,
 * including recognition of partial dates and intuitive defaulting of omitted components.
 * 
 * <p> Ranges are 'from' a date, 'to' a date, with either or both being open-ended.
 * Inclusivity vs. exclusivity of range ends are supported via the keywords:
 * <ul>
 * <li> {@code from} - inclusive 'from' part of date range
 * <li> {@code to} - inclusive 'to' part of date range
 * <li> {@code after} - exclusive 'from' part of date range
 * <li> {@code before} - exclusive 'to' part of date range
 * </ul>
 * 
 * <p> The following formats are supported (excluding enclosing double quotes):
 * <table border="1">
 * 	<tr>
 *    <th>Format</th>
 *    <th>Description</th>
 *    <th>Examples</th>
 *  </tr>
 *  <tr>
 *    <td>{@code date}</td>
 *    <td>Range from the start to the end of the date or partial date given</td>
 *    <td>{@code 2/mar/2013}<br/>
 *        2013/Mar/02 00:00:00 to 2013/Mar/02 23:59:59<br/>
 *        <br/>
 *        {@code mar/2013}<br/>
 *        2013/Mar/01 00:00:00 to 2013/Mar/31 23:59:59<br/>
 *        <br/>
 *        {@code 2013}<br/>
 *        2013/Jan/01 00:00:00 to 2013/Dec/31 23:59:59
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code date to date}<br/><br/>
 *        {@code date, to date}</br>
 *        {@code from date to date}</br>
 *        {@code from date, to date}</td>
 *    <td>Range from the start of the first given date or partial date,
 *        to the end of the second given date or partial date.
 *        <br/>The {@code from} keyword and punctuation are optional here.
 *        <br/>Punctuation accepted are comma (,), semicolon (;), and colon (:).
 *    </td>
 *    <td>{@code 2/mar/2013 to 5/August/2013}<br/>
 *        2013/Mar/02 00:00:00 to 2013/Aug/05 23:59:59<br/>
 *        <br/>
 *        {@code mar/2013 to aug/2013}<br/>
 *        2013/Mar/01 00:00:00 to 2013/Aug/31 23:59:59<br/>
 *        <br/>
 *        {@code 2013 to 2015}<br/>
 *        2013/Jan/01 00:00:00 to 2015/Dec/31 23:59:59
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code after date before date}<br/><br/>
 *        {@code after date, before date}</br>
 *    <td>Range from the instant after the end of the first given date or partial date,
 *        to the instant before the start of the second given date or partial date.
 *        <br/>Punctuation is optional; accepted are comma (,), semicolon (;), and colon (:).
 *    </td>
 *    <td>{@code after 2/mar/2013, before 5/August/2013}<br/>
 *        2013/Mar/03 00:00:00 to 2013/Aug/04 23:59:59<br/>
 *        <br/>
 *        {@code after mar/2013 before aug/2013}<br/>
 *        2013/Apr/01 00:00:00 to 2013/July/31 23:59:59<br/>
 *        <br/>
 *        {@code after 2013; before 2015}<br/>
 *        2014/Jan/01 00:00:00 to 2014/Dec/31 23:59:59
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code from date before date}<br/><br/>
 *        {@code after date to date}</br>
 *    <td>Keywords may be mixed.
 *    </td>
 *    <td>{@code after 2/mar/2013, to 5/August/2013}<br/>
 *        2013/Mar/03 00:00:00 to 2013/Aug/05 23:59:59<br/>
 *        <br/>
 *        {@code mar/2013 before aug/2013}<br/>
 *        2013/Mar/01 00:00:00 to 2013/July/31 23:59:59<br/>
 *        <br/>
 *        {@code from 2013; before 2015}<br/>
 *        2013/Jan/01 00:00:00 to 2014/Dec/31 23:59:59
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code from date}</td>
 *    <td>Open ended range from the start of the date given to positive infinity</td>
 *    <td>{@code from 2/mar/2013}<br/>
 *        2013/Mar/02 00:00:00 to null<br/>
 *        <br/>
 *        {@code from mar/2013}<br/>
 *        2013/Mar/01 00:00:00 to null<br/>
 *        <br/>
 *        {@code from 2013}<br/>
 *        2013/Jan/01 00:00:00 to null
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code to date}</td>
 *    <td>Open ended range from negative infinity to the end of the date given</td>
 *    <td>{@code to 2/mar/2013}<br/>
 *        null to 2013/Mar/02 23:59:59<br/>
 *        <br/>
 *        {@code to mar/2013}<br/>
 *        null to 2013/Mar/31 23:59:59<br/>
 *        <br/>
 *        {@code to 2013}<br/>
 *        null to 2013/Dec/31 23:59:59
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code after date}</td>
 *    <td>Open ended range from the instant after the end of the date given to positive infinity</td>
 *    <td>{@code after 2/mar/2013}<br/>
 *        2013/Mar/03 00:00:00 to null<br/>
 *        <br/>
 *        {@code after mar/2013}<br/>
 *        2013/Apr/01 00:00:00 to null<br/>
 *        <br/>
 *        {@code after 2013}<br/>
 *        2014/Jan/01 00:00:00 to null
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>{@code before date}</td>
 *    <td>Open ended range from negative infinity to the instant before the start of the date given</td>
 *    <td>{@code before 2/mar/2013}<br/>
 *        null to 2013/Mar/01 23:59:59<br/>
 *        <br/>
 *        {@code to mar/2013}<br/>
 *        null to 2013/Feb/28 23:59:59<br/>
 *        <br/>
 *        {@code to 2013}<br/>
 *        null to 2012/Dec/31 23:59:59
 *    </td>
 *  </tr>
 * </table>
 * 
 * <p> Internally uses {@link FlexibleDateFormat} to parse and format individual dates.
 * Ddate components may be separated by hyphens, slashes or spaces.
 * Months may be numerals or names. Years may be two or four digits.
 * Date components may be omitted, provided an unambiguous date can be derived.
 * Omitted components are defaulted based on the following rules:
 * <ul>
 * <li> Years are defaulted according to the {@link YearRangeBias} (default {@link YearRangeBias#NEAREST}
 * <li> Months are defaulted to the start or end of the year for the 'from' and 'to' components of the range, respectively.
 * <li> Days are defaulted to the start or end of the month for the 'from' and 'to' components of the range, respectively.
 * <li> Time-of-date is defaulted to the start of end of the day for the 'from' and 'to' components of the range, respectively.
 * </ul>
 * 
 * <p> <u>Consistency</u><br/>
 * Generates ranges are always "consistent":
 * if both {@code from} and {@code to} dates are defined, then {@code from <= to}.
 * This applies when {@code after} and {@code before} are used instead.
 *
 * <p> <u>Inheritance</u><br/>
 * Where a {@code from} (or {@code after}) date and a {@code to} (or {@code before}) date are specified,
 * but only one has a specified year, the date with undefined year "inherits" the year of the other.
 * The inherited year is modified forward or backward by one year in order to maintain a consistent range.
 * 
 * <p> For example the following dates in the first column are interpreted the same as in the second column:
 * <table border="1">
 * <tr><th>Example</th><th>Interpreted As</th></tr>
 * <tr><td>{@code mar to aug/2009}</td><td>{@code mar/2009 to aug/2009}</td></tr>
 * <tr><td>{@code dec to aug/2009}</td><td>{@code dec/2008 to aug/2009}</td></tr>
 * <tr><td>{@code mar/2009 to aug}</td><td>{@code mar/2009 to aug/2009}</td></tr>
 * <tr><td>{@code dec/2008 to aug}</td><td>{@code dec/2008 to aug/2009}</td></tr>
 * </table>
 * 
 * <p> <u>Year Range Bias</u><br/>
 * Years are defaulted according to the configured year bias when not supplied in the input.
 * Year inheritance supercedes the year bias, so the year bias is only applied where
 * no year is given.
 * 
 * <p>Warning: instances of this class are not thread safe. Using the same instance
 * to parse date between multiple threads simultaneously will produce unpredictable results.
 * @author Malcolm Lett
 */
public class FlexibleDateRangeFormat extends Format {
	private static final long serialVersionUID = 1L;
	private static final Logger log = Logger.getLogger(FlexibleDateRangeFormat.class);
	private static final Pattern MAIN_PART_DELIMETER = Pattern.compile("(from|to|before|after|[,;:])");
	private static final Pattern UNWANTED_MAIN_PART_DELIMETER = Pattern.compile("[,;:]");
	private static final YearRangeBias DEFAULT_YEAR_RANGE_BIAS = YearRangeBias.NEAREST;

	private final DatePattern pattern;
	private final Locale locale;
	private final DateFormatSymbols formatSymbols;
	private final YearRangeBias yearBias;
	private final FlexibleDateFormat startDateFormat;
	private final FlexibleDateFormat endDateFormat;
	private final List<RangePostProcessor> postProcessors;
	
	/**
	 * Gets a default date range formatter that uses the default MEDIUM style for date.
	 * <p> Same as returned by {@link #getDateInstance()}.
	 * <p> <i>Note:</i> this method is subject to change. When time is supported
	 * this method will return a date/time formatter instead.
	 * @return a date formatter.
	 */
	public static FlexibleDateRangeFormat getDefaultInstance() {
		return getDateInstance();
	}
	
	/**
	 * Gets a date range formatter with the given pattern
	 * and the default date format symbols for the default locale.
	 * @param pattern the pattern describing the date and time format
	 * @throws NullPointerException if the given pattern is null
	 */
	public static FlexibleDateRangeFormat getPatternInstance(String pattern) {
		return new FlexibleDateRangeFormat(new DatePattern(pattern), Locale.getDefault());
	}
	
	/**
	 * Gets a date range formatter with the given pattern
	 * and the default date format symbols for the given locale.
	 * @param pattern the pattern describing the date and time format
	 * @param locale the locale whose date format symbols should be used   
	 * @throws NullPointerException if the given pattern or locale is null
	 */
	public static FlexibleDateRangeFormat getPatternInstance(String pattern, Locale locale) {
		return new FlexibleDateRangeFormat(new DatePattern(pattern), locale, new DateFormatSymbols(locale));
	}
	
	/**
	 * Gets a date range formatter with the given pattern and date format symbols.
	 * @param pattern the pattern describing the date and time format
	 * @param formatSymbols the date format symbols to be used for formatting   
	 * @throws NullPointerException if the given pattern or formatSymbols is null
	 */
	public static FlexibleDateRangeFormat getPatternInstance(String pattern, DateFormatSymbols formatSymbols) {
		return new FlexibleDateRangeFormat(new DatePattern(pattern), Locale.getDefault(), formatSymbols);
	}
	
	/**
	 * Gets a default date formatter that uses the default MEDIUM style for date.
	 * @return a date formatter.
	 */
	public static FlexibleDateRangeFormat getDateInstance() {
		return new FlexibleDateRangeFormat(DatePattern.getDateInstance(), Locale.getDefault());
	}
	
	/**
	 * Gets a date range formatter with the given formatting style for the default locale.
	 * @param style the given formatting style, defined within the {@link DateFormat} class.
	 *        For example, {@link DateFormat#SHORT} for "M/d/yy" in the US locale.
	 * @throws NullPointerException if the given style is null
	 * @return a date formatter.
	 */
	public static FlexibleDateRangeFormat getDateInstance(int style) {
		return getDateInstance(style, Locale.getDefault());
	}
	
	/**
	 * Gets a date range formatter with the given formatting style for the given locale.
	 * @param style the given formatting style, defined within the {@link DateFormat} class.
	 *        For example, {@link DateFormat#SHORT} for "M/d/yy" in the US locale.
	 * @param locale the given locale
	 * @throws NullPointerException if the given style or locale is null
	 * @return a date formatter.
	 */
	public static FlexibleDateRangeFormat getDateInstance(int style, Locale locale) {
		return new FlexibleDateRangeFormat(DatePattern.getDateInstance(style, locale), locale);
	}
	
	/**
	 * Returns an array of all locales for which the
	 * <code>get*Instance</code> methods of this class can return
	 * localized instances.
	 * The returned array represents the union of locales supported by the Java
	 * runtime and by installed
	 * {@link java.text.spi.DateFormatProvider DateFormatProvider} implementations.
	 * It must contain at least a <code>Locale</code> instance equal to
	 * {@link java.util.Locale#US Locale.US}.
	 *
	 * @return An array of locales for which localized
	 *         <code>FlexibleDateRangeFormat</code> instances are available.
	 */
	public static Locale[] getAvailableLocales() {
		return DateFormat.getAvailableLocales();
	}
	
	// package-private intermediate constructor
	FlexibleDateRangeFormat(DatePattern pattern, Locale locale) {
		this(pattern, locale, new DateFormatSymbols(locale), DEFAULT_YEAR_RANGE_BIAS);
	}
	
	// package-private intermediate constructor
	FlexibleDateRangeFormat(DatePattern pattern, Locale locale, DateFormatSymbols formatSymbols) {
		this(pattern, locale, formatSymbols, DEFAULT_YEAR_RANGE_BIAS);
	}
	
	/**
	 * Constructs a {@code FlexibleDateRangeFormat} using the given pattern,
	 * locale and date format symbols
	 * @param pattern the pattern describing the date and time format
	 * @param locale the locale whose date format symbols should be used   
	 * @param formatSymbols the date format symbols to be used for formatting   
	 * @param yearBias the year bias applied when defaulting years   
	 * @throws NullPointerException if the given pattern, locale, formatSymbols, or yearBias is null
	 * @throws IllegalArgumentException if the given pattern is invalid 
	 */
	private FlexibleDateRangeFormat(DatePattern pattern, Locale locale, DateFormatSymbols formatSymbols, YearRangeBias yearBias) {
		if (pattern == null) {
			throw new NullPointerException("pattern must be specified");
		}
		if (locale == null) {
			throw new NullPointerException("locale must be specified");
		}
		if (formatSymbols == null) {
			throw new NullPointerException("formatSymbols must be specified");
		}
		if (yearBias == null) {
			throw new NullPointerException("yearBias must be specified");
		}
		
		this.pattern = pattern;
		this.locale = locale;
		this.formatSymbols = formatSymbols;
		this.yearBias = yearBias;
		this.startDateFormat = new FlexibleDateFormat(pattern, locale, formatSymbols, yearDateBiasFor(yearBias), Bias.START);
		this.endDateFormat   = new FlexibleDateFormat(pattern, locale, formatSymbols, yearDateBiasFor(yearBias), Bias.END);
		
		this.postProcessors = new ArrayList<RangePostProcessor>();
		this.postProcessors.add(inheritDefinedYear());
		this.postProcessors.add(defaultYearsIfUndefined(yearBias));
	}
	
	/**
	 * Gets a new instance that is a clone of this one plus it is configured
	 * with the supplied year bias.
	 * @param yearBias bias used when defaulting years
	 * @return the new instance
	 */
	public FlexibleDateRangeFormat withYearBias(YearRangeBias yearBias) {
		return new FlexibleDateRangeFormat(pattern, locale, formatSymbols, yearBias);
	}
	
	/**
	 * Gets the pattern this date format is working to.
	 */
	@Override
	public String toString() {
		return startDateFormat.toString();
	}

	/**
	 * Gets whether partial dates are formatted partially or fully.
	 */
	public boolean isFormatPartially() {
		return startDateFormat.isFormatPartially();
	}

	/**
	 * Sets whether partial dates are formatted partially or fully.
	 * This applies to the least-significant components of the date only. 
	 * @param formatPartially
	 */
	public void setFormatPartially(boolean formatPartially) {
		startDateFormat.setFormatPartially(formatPartially);
		endDateFormat.setFormatPartially(formatPartially);
	}
	
	/**
	 * Gets the configured year bias.
	 */
	public YearRangeBias getYearBias() {
		return yearBias;
	}
	
	/**
	 * Sets the time zone in use when formatting and parsing.
	 * @param zone the given new time zone.
	 */
	public void setTimeZone(TimeZone zone) {
		startDateFormat.setTimeZone(zone);
		endDateFormat.setTimeZone(zone);
	}
	
	/**
	 * Gets the time zone.
	 * @return the time zone in use when formatting and parsing.
	 */
	public TimeZone getTimeZone() {
		return startDateFormat.getTimeZone();
	}
	
	/**
	 * Overrides hashcode
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((endDateFormat == null) ? 0 : endDateFormat.hashCode());
		result = prime * result
				+ ((startDateFormat == null) ? 0 : startDateFormat.hashCode());
		result = prime * result
				+ ((yearBias == null) ? 0 : yearBias.hashCode());
		return result;
	}

	/**
	 * Overrides equals
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof FlexibleDateRangeFormat)) {
			return false;
		}
		FlexibleDateRangeFormat other = (FlexibleDateRangeFormat) obj;
		if (endDateFormat == null) {
			if (other.endDateFormat != null) {
				return false;
			}
		} else if (!endDateFormat.equals(other.endDateFormat)) {
			return false;
		}
		if (startDateFormat == null) {
			if (other.startDateFormat != null) {
				return false;
			}
		} else if (!startDateFormat.equals(other.startDateFormat)) {
			return false;
		}
		if (yearBias != other.yearBias) {
			return false;
		}
		return true;
	}

	/**
	 * Overrides Cloneable.
	 * Not a complete clone implementation.
	 * Probably best not to use.
	 */
	@Override
	public FlexibleDateRangeFormat clone() {
		FlexibleDateRangeFormat other = (FlexibleDateRangeFormat) super.clone();
		return other;
		
		// doesn't work because all these fields are declared 'final' 
//		FlexibleDateRangeFormat other = (FlexibleDateRangeFormat) super.clone();
//		other.endDateFormat = this.endDateFormat.clone();
//		other.startDateFormat = this.startDateFormat.clone();
//		other.locale = this.locale;
//		other.pattern = this.pattern;
//		other.postProcessors = this.postProcessors;
//		other.yearBias = this.yearBias;
//		return other;
	}
	
	/**
	 * Formats the date range according to the level at which it is defined.
	 * @param dateRange the date range to format, may be null
	 * @return the string representation of the date range
	 * @throws NullPointerException if range is null
	 */
	public String format(DateRange dateRange) {
		return format(dateRange, new StringBuffer(), new FieldPosition(DateFormat.YEAR_FIELD)).toString();
	}
	
	/**
	 * Formats a {@link DateRange} object into a string.
	 * @param obj must be a {@link DateRange}
	 * @param toAppendTo the string buffer for the returning time string.
	 * @param fieldPosition keeps track of the position of the first occurrence of
	 * the field within the returned string.
	 * On input: an alignment field, if desired.
	 * On output: the offsets of the alignment field.
	 * For example, given a date text "1996.07.10 AD",
	 * if the given fieldPosition is DateFormat.YEAR_FIELD, the
	 * begin index and end index of fieldPosition will be set to
	 * 0 and 4, respectively.
	 * @return the string buffer passed in as toAppendTo, with formatted text appended.
	 * @see java.text.Format
	 */
	@Override
	public StringBuffer format(Object obj, StringBuffer toAppendTo, FieldPosition fieldPosition) {
		if (obj instanceof DateRange) {
			return format( (Date)obj, toAppendTo, fieldPosition );
		}
		else {
			throw new IllegalArgumentException("Cannot format given Object as a Date");
		}
	}
	
	/**
	 * Formats the date range according to the level at which it is defined.
	 * @param dateRange the date range to format, may be null
	 * @param toAppendTo the string buffer for the returning time string.
	 * @param fieldPosition keeps track of the position of the first occurrence of
	 * the field within the returned string.
	 * On input: an alignment field, if desired.
	 * On output: the offsets of the alignment field.
	 * For example, given a date text "1996.07.10 AD",
	 * if the given fieldPosition is DateFormat.YEAR_FIELD, the
	 * begin index and end index of fieldPosition will be set to
	 * 0 and 4, respectively.
	 * @see java.text.Format
	 * @return the string buffer passed in as toAppendTo, with formatted text appended.
	 * @throws NullPointerException if range is null
	 */
	public StringBuffer format(DateRange dateRange, StringBuffer toAppendTo, FieldPosition fieldPosition) {
		DateRangeBuilder range = DateRangeBuilder.forFormatOf(dateRange);
		
		DefinedLevel finerLevel = DefinedLevel.lessSignificantOf(
				(range.getFrom() == null) ? DefinedLevel.YEAR : range.getFrom().definedTo(),
				(range.getTo()   == null) ? DefinedLevel.YEAR : range.getTo().definedTo());
		
		// from/to - contracted  (from equals to)
		if (range.getFrom() != null && range.getTo() != null &&
				range.getFrom().definedTo() == range.getTo().definedTo() &&
				range.getFrom().compareTo(range.getTo(), finerLevel) == 0) {
			toAppendTo.append(startDateFormat.format(range.getFrom()));
		}
		
		// from/to - uncontracted
		else if (range.getFrom() != null && range.getTo() != null) {
			toAppendTo.append(startDateFormat.format(range.getFrom().defineTo(finerLevel)));
			toAppendTo.append(" to ");
			toAppendTo.append(endDateFormat.format(range.getTo().defineTo(finerLevel)));
		}
		
		// after only
		else if (range.getFrom() != null) {
			toAppendTo.append("from ");
			toAppendTo.append(startDateFormat.format(range.getFrom()));
		}

		// before only
		else if (range.getTo() != null) {
			toAppendTo.append("to ");
			toAppendTo.append(endDateFormat.format(range.getTo()));
		}
		return toAppendTo;
	}

	/**
	 * Parses the supplied text into a date range.
	 * @param source the string representation of the date range to be parsed
	 * @return the parsed date range object
	 * @throws ParseException on any parse errors
	 */
	public DateRange parse(String source) throws ParseException {
		return parseInternal(source, new ParsePosition(0));
	}

	/**
	 * Parses the supplied text into a date range.
	 * @param text the string representation of the date range to be parsed
	 * @param pos   On input, the position at which to start parsing; on
	 *              output, the position at which parsing terminated, or the
	 *              start position if the parse failed.
	 * @return A {@code DateRange}, or {@code null} if the input could not be parsed
	 */
	public DateRange parse(String text, ParsePosition pos) {
		try {
			return parseInternal(text, pos);
		} catch (ParseException dropped) {
			return null;
		}
	}
	
	/**
	 * Parses text from the beginning of the given string to produce an object.
	 * The method may not use the entire text of the given string.
	 *
	 * @param source A <code>String</code> whose beginning should be parsed.
	 * @return An <code>Object</code> parsed from the string.
	 * @exception ParseException if the beginning of the specified string
	 *            cannot be parsed.
	 */
	@Override
	public Object parseObject(String source) throws ParseException {
		return parseInternal(source, new ParsePosition(0));
	}
	
	/**
	 * Parses text from a string to produce a {@code DateRange}.
	 * <p>
	 * The method attempts to parse text starting at the index given by {@code pos}.
	 * If parsing succeeds, then the index of {@code pos} is updated
	 * to the index after the last character used (parsing does not necessarily
	 * use all characters up to the end of the string), and the parsed
	 * date is returned. The updated {@code pos} can be used to
	 * indicate the starting point for the next call to this method.
	 * If an error occurs, then the index of {@code pos} is not
	 * changed, the error index of {@code pos} is set to the index of
	 * the character where the error occurred, and null is returned.
	 * <p>
	 * See the {@link #parse(String, ParsePosition)} method for more information
	 * on date parsing.
	 *
	 * @param source A <code>String</code>, part of which should be parsed.
	 * @param pos A <code>ParsePosition</code> object with index and error
	 *            index information as described above.
	 * @return A <code>DateRange</code> parsed from the string. In case of
	 *         error, returns null.
	 * @exception NullPointerException if <code>pos</code> is null.
	 */
	@Override
	public Object parseObject(String source, ParsePosition pos) {
		try {
			return parseInternal(source, pos);
		} catch (ParseException dropped) {
			return null;
		}
	}
	
	/**
	 * Parses the supplied text into a date range.
	 * @param source the string representation of the date to be parsed
	 * @param pos   On input, the position at which to start parsing; on
	 *              output or exception, the position at which parsing terminated, or the
	 *              start position if the parse failed.
	 * @return the date if successfully parsed
	 * @throws ParseException on any parse errors
	 */
	protected DateRange parseInternal(String source, ParsePosition pos) throws ParseException {
		if (source == null) {
			throw new NullPointerException("cannot parse null strings");
		}
		source = source.trim();
		if ("".equals(source.trim())) {
			throw new ParseException("cannot parse empty strings",0);
		}
		
		// standardise source
		source = source.toLowerCase();
		source = source.replace("/", "-");
		source = source.replace("\\", "-");
		
		// split into main parts
		List<String> mainParts = splitKeepingDelimeters(source, MAIN_PART_DELIMETER);
		mainParts = cleanupMainParts(mainParts);
		
		// parse main parts
		ParseResult parseResult = new ParseResult();
		for (String mainPart: mainParts) {
			parseMainPart(mainPart, parseResult);
		}
		
		// final grammar business rules
		if (parseResult.standaloneFrom != null && parseResult.from != null) {
			throw new ParseException("standalone date and 'from' not allowed together",0);
		}
		if (parseResult.standaloneFrom != null && parseResult.after != null) {
			throw new ParseException("standalone date and 'after' not allowed together",0);
		}
		if (parseResult.from != null && parseResult.after != null) {
			throw new ParseException("'from' and 'after' not allowed together",0);
		}
		if (parseResult.to != null && parseResult.before != null) {
			throw new ParseException("'to' and 'before' not allowed together",0);
		}
		
		// construct result and apply post-processing
		DateRangeBuilder range = DateRangeBuilder.fromTo(parseResult.getFrom(), parseResult.getTo());
		for (RangePostProcessor postProcessor: postProcessors) {
			range = postProcessor.process(range);
		}
		
		// validate consistency
		if (range.getFrom() != null && range.getTo() != null) {
			if (range.getFrom().compareTo(range.getTo()) == 0) {
				throw new ParseException("range rejects all dates",0);
			}
			else if (range.getFrom().compareTo(range.getTo()) > 0) {
				throw new ParseException("start date is after end date",0);
			}
		}
		log.debug("Parsed ["+source+"] as "+range);
		return range.asDateRange();
	}
	
	/**
	 * Parses the main part given and appends it to the parse result.
	 * Performs error checking, including the following business rules:
	 * <ul>
	 * <li> each keyword is only allowed once.
	 * <li> only one of 'from' and 'after' are allowed.
	 * <li> only one of 'to' and 'before' are allowed.
	 * <li> unrecognised keywords are assumed part of the date and the date is attempted to be parsed
	 * </ul>
	 * @param mainPart
	 * @param parseResult
	 */
	private void parseMainPart(String mainPart, ParseResult parseResult) throws ParseException {
		log.debug(" parsing part ["+mainPart+"]");
		// region with valid keyword
		if (mainPart.startsWith("from")) {
			if (parseResult.from != null) {
				throw new ParseException("'from' not allowed more than once",0);
			}
			String dateStr = mainPart.substring("from".length()).trim();
			parseResult.from = startDateFormat.parse(dateStr);
		}
		else if (mainPart.startsWith("to")) {
			if (parseResult.to != null) {
				throw new ParseException("'to' not allowed more than once",0);
			}
			String dateStr = mainPart.substring("to".length()).trim();
			parseResult.to = endDateFormat.parse(dateStr);
		}
		else if (mainPart.startsWith("after")) {
			if (parseResult.after != null) {
				throw new ParseException("'after' not allowed more than once",0);
			}
			String dateStr = mainPart.substring("after".length()).trim();
			parseResult.after = endDateFormat.parse(dateStr);
		}
		else if (mainPart.startsWith("before")) {
			if (parseResult.before != null) {
				throw new ParseException("'before' not allowed more than once",0);
			}
			String dateStr = mainPart.substring("before".length()).trim();
			parseResult.before = startDateFormat.parse(dateStr);
		}
		
		// first region without keyword (standalone or 'from')
		else if (parseResult.isEmpty()) {
			parseResult.standaloneFrom = startDateFormat.parse(mainPart);
			parseResult.standaloneTo   = endDateFormat.parse(mainPart);
		}
		
		// invalid region
		else {
			throw new ParseException("Date missing keyword '"+mainPart+"'", 0);
		}
	}
	
	/**
	 * Removes unwanted commas etc., trims, and removes empty parts.
	 * @param parts
	 * @return main parts after exclusions
	 */
	private List<String> cleanupMainParts(List<String> parts) {
		List<String> result = new ArrayList<String>();
		for (String part: parts) {
			part = UNWANTED_MAIN_PART_DELIMETER.matcher(part).replaceAll("");
			part = part.trim();
			if (part.length() > 0) {
				result.add(part);
			}
		}
		return result;
	}
	
	/**
	 * Split that keeps the delimiters at the start of each region.
	 * Removes empty regions.
	 * Returns an empty list if no non-empty regions.
	 * @param input
	 * @param regex
	 * @return list of non-excluded string components
	 */
	static List<String> splitKeepingDelimeters(String input, Pattern regex) {
		List<String> regions = new ArrayList<String>();
		Matcher m = regex.matcher(input);
		int lastStart = 0;
		while (m.find()) {
			String region = input.substring(lastStart, m.start());
			if (region.length() > 0) {
				regions.add(region);
			}
			lastStart = m.start();
		}
		
		// add last region
		String region = input.substring(lastStart);
		if (region.length() > 0) {
			regions.add(region);
		}
		return regions;
	}

	/**
	 * Internal support method to check whether the year is defined.
	 */
	protected static boolean yearDefined(PartialDate date) {
		return date.isDefinedAt(DefinedLevel.YEAR);
	}
	
	/**
	 * Internal support method for quick access to method within code.
	 */
	// Could equally implement using org.apache.commons.lang3.DateUtils.toCalendar(Date), but that does the same thing anyway.
	// TODO: consider whether need to apply timezone smarts as in java.text.SimpleDateFormat.initializeCalendar(Locale)
	private static Calendar calendarOf(Date date) {
		Calendar calendar = Calendar.getInstance();
		calendar.setTime(date);
		return calendar;
	}

	/**
	 * Determines the year bias that is best suited to the given year range bias.
	 * Note: that the bias applied when parsing the individual dates of the range
	 * are only the first step in defaulting the years.
	 * For example, in the case of {@link YearRangeBias#NEAREST} there is no
	 * need for fancy defaulting in the first stage.
	 */
	private static YearBias yearDateBiasFor(YearRangeBias yearRangeBias) {
		switch (yearRangeBias) {
			case NEAREST:         return YearBias.CURRENT;
			case CURRENT:         return YearBias.CURRENT;
			case STRICTLY_PAST:   return YearBias.PAST;
			case PAST:            return YearBias.PAST;
			case FUTURE:          return YearBias.FUTURE;
			case STRICTLY_FUTURE: return YearBias.FUTURE;
			default: throw new UnsupportedOperationException("Unexpected year bias "+yearRangeBias);
		}
	}
	
	/**
	 * Gets a post processor that applies year inheritance
	 * where exactly one date in a range has a defined year.
	 */
	protected static RangePostProcessor inheritDefinedYear() {
		return new RangePostProcessor() {
			@Override
			public DateRangeBuilder process(DateRangeBuilder range) throws ParseException {
				if (range.getFrom() != null && range.getTo() != null) {
					boolean fromYearDefined = yearDefined(range.getFrom());
					boolean toYearDefined = yearDefined(range.getTo());
					if (fromYearDefined && !toYearDefined) {
						PartialDateBuilder from = new PartialDateBuilder(range.getFrom(), Bias.START);
						PartialDateBuilder to = new PartialDateBuilder(range.getTo(), Bias.END);
						to.set(Calendar.YEAR, from.get(Calendar.YEAR));
						incrementToYearIfInconsistent(from, to);
						range = DateRangeBuilder.fromTo(from.asPartialDate(), to.asPartialDate());
					}
					else if (!fromYearDefined && toYearDefined) {
						PartialDateBuilder from = new PartialDateBuilder(range.getFrom(), Bias.START);
						PartialDateBuilder to = new PartialDateBuilder(range.getTo(), Bias.END);
						from.set(Calendar.YEAR, to.get(Calendar.YEAR));
						decrementFromYearIfInconsistent(from, to);
						range = DateRangeBuilder.fromTo(from.asPartialDate(), to.asPartialDate());
					}
				}
				return range;
			}
			
			private void decrementFromYearIfInconsistent(PartialDateBuilder from, PartialDateBuilder to) {
				if (from.compareTo(to) > 0) {
					from.add(Calendar.YEAR, -1);
				}
			}

			private void incrementToYearIfInconsistent(PartialDateBuilder from, PartialDateBuilder to) {
				if (from.compareTo(to) > 0) {
					to.add(Calendar.YEAR, +1);
				}
			}
		};
	}

	/**
	 * Gets a post processor that defaults the years
	 * of the dates in the range, according to the given bias,
	 * but only if both dates in the range have undefined year.
	 * 
	 * <p> Note that years will already have been defaulted according
	 * to the {@code YearBias} configured earlier.
	 * This processor has the responsibility of ensuring that year defaulting
	 * results in a "consistent" range, and applying further defaulting where necessary.
	 */
	protected static RangePostProcessor defaultYearsIfUndefined(YearRangeBias yearBias) {
		switch (yearBias) {
			// Both will be given the current year, which could result in an inconsistent range.
			// No further defaulting is done.
			// Rely on subsequent validation to assert that the range is consistent. 
			case CURRENT: return noop();

			// adjust years based on mid-point
			case NEAREST: return adjustMidpointToNearestIfInconsistent();
				
			// 'to' year will have been defaulted correctly,
			// fix the 'from' year by decrementing if the range is inconsistent
			case FUTURE: return decrementFromYearIfInconsisent();
				
			// 'from' year will have been defaulted correctly,
			// fix the 'to' year by incrementing if the range is inconsistent
			case PAST: return incrementToYearIfInconsisent();
				
			// 'from' year will have been defaulted correctly,
			// fix the 'to' year by incrementing if the range is inconsistent
			case STRICTLY_FUTURE: return incrementToYearIfInconsisent();
				
			// 'to' year will have been defaulted correctly,
			// fix the 'from' year by decrementing if the range is inconsistent
			case STRICTLY_PAST: return decrementFromYearIfInconsisent();
				
			default: throw new UnsupportedOperationException("Unexpected "+YearRangeBias.class.getSimpleName()+" "+yearBias);
		}
	}
	
	/**
	 * If both years are undefined, then sets the years according to {@link YearRangeBias#NEAREST}:
	 * <ol>
	 * <li> adjusts 'from' year so that range is consistent, then
	 * <li> adjusts 'from' and 'to' years so that the range mid-point is closest
	 *      to the current day.
	 * </ol>
	 */
	protected static RangePostProcessor adjustMidpointToNearestIfInconsistent() {
		return new RangePostProcessor() {
			private RangePostProcessor makeConsistent = decrementFromYearIfInconsisent();
			private DatePostProcessor midYearProcessor = FlexibleDateFormat.nearestYear();
			@Override
			public DateRangeBuilder process(DateRangeBuilder range) throws ParseException {
				if (range.getFrom() != null && range.getTo() != null) {
					if (!yearDefined(range.getFrom()) && !yearDefined(range.getTo())) {
						// make consistent
						range = makeConsistent.process(range);

						// calculate target mid point
						Date midDate = midOf(range.getFrom(), range.getTo());
						midDate = midYearProcessor.process(midDate);
						PartialDateBuilder mid = new PartialDateBuilder(midDate, null);
						
						// set years of range
						PartialDateBuilder from = new PartialDateBuilder(range.getFrom(), Bias.START);
						PartialDateBuilder to = new PartialDateBuilder(range.getTo(), Bias.END);
						from.set(Calendar.YEAR, mid.get(Calendar.YEAR));
						to.set(Calendar.YEAR,   mid.get(Calendar.YEAR));
						
						decrementFromYearIfInconsistent(from, mid);
						incrementToYearIfInconsistent(mid, to);
						
						return DateRangeBuilder.fromTo(from.asPartialDate(), to.asPartialDate());
					}
				}
				return range;
			}

			private void decrementFromYearIfInconsistent(PartialDateBuilder from, PartialDateBuilder to) {
				if (from.compareTo(to) > 0) {
					from.add(Calendar.YEAR, -1);
				}
			}

			private void incrementToYearIfInconsistent(PartialDateBuilder from, PartialDateBuilder to) {
				if (from.compareTo(to) > 0) {
					to.add(Calendar.YEAR, +1);
				}
			}
			
			// note: accurate to about 2ms, which is plenty for the application
			private Date midOf(Date one, Date two) {
				long oneMillis = one.getTime();
				long twoMillis = two.getTime();
				Date mid = new Date(oneMillis/2 + twoMillis/2); // avoids java long overflow
				
				// safety: ensure rounding errors don't cause problems if one==two
				if (mid.compareTo(one) < 0) {
					mid = one;
				}
				return mid;
			}
		};
	}
	
	/**
	 * If the date range is in an inconsistent state, then assumes that the 'to'
	 * year is correct and adjusts the 'from' year.
	 * The 'from' year will be set equal to the 'to' year or one before as needed. 
	 */
	protected static RangePostProcessor decrementFromYearIfInconsisent() {
		return new RangePostProcessor() {
			@Override
			public DateRangeBuilder process(DateRangeBuilder range) throws ParseException {
				if (range.getFrom() != null && range.getTo() != null && range.getFrom().compareTo(range.getTo()) > 0) {
					if (!yearDefined(range.getFrom()) && !yearDefined(range.getTo())) {
						PartialDateBuilder from = new PartialDateBuilder(range.getFrom(), Bias.START);
						PartialDateBuilder to = new PartialDateBuilder(range.getTo(), Bias.END);
						from.set(Calendar.YEAR, to.get(Calendar.YEAR));
						if (from.compareTo(to) > 0 ){
							from.add(Calendar.YEAR, -1);
						}
						return DateRangeBuilder.fromTo(from.asPartialDate(), to.asPartialDate());
					}
				}
				return range;
			}
		};
	}

	/**
	 * If the date range is in an inconsistent state, then assumes that the 'from'
	 * year is correct and adjusts the 'to' year. 
	 * The 'to' year will be set equal to the 'from' year or one after as needed. 
	 */
	protected static RangePostProcessor incrementToYearIfInconsisent() {
		return new RangePostProcessor() {
			@Override
			public DateRangeBuilder process(DateRangeBuilder range) throws ParseException {
				if (range.getFrom() != null && range.getTo() != null && range.getFrom().compareTo(range.getTo()) > 0) {
					if (!yearDefined(range.getFrom()) && !yearDefined(range.getTo())) {
						PartialDateBuilder from = new PartialDateBuilder(range.getFrom(), Bias.START);
						PartialDateBuilder to = new PartialDateBuilder(range.getTo(), Bias.END);
						to.set(Calendar.YEAR, from.get(Calendar.YEAR));
						if (from.compareTo(to) > 0 ){
							to.add(Calendar.YEAR, +1);
						}
						return DateRangeBuilder.fromTo(from.asPartialDate(), to.asPartialDate());
					}
				}
				return range;
			}
		};
	}
	
	/**
	 * Does nothing.
	 */
	protected static RangePostProcessor noop() {
		return new RangePostProcessor() {
			@Override
			public DateRangeBuilder process(DateRangeBuilder range) {
				return range;
			}
		};
	}
	
	/**
	 * Internal support class used to store the parsed result as it is built up.
	 */
	private static class ParseResult {
		private PartialDate standaloneFrom;
		private PartialDate standaloneTo; // might not be used, even if set
		private PartialDate from;
		private PartialDate to;
		private PartialDate after;
		private PartialDate before;
		
		// indicates that no regions have been parsed yet
		public boolean isEmpty() {
			return (standaloneFrom == null) && (standaloneTo == null) &&
					(from == null) && (to == null) && (after == null) && (before == null);
		}
		
		public PartialDate getFrom() {
			if (from != null) {
				return from;
			}
			else if (after != null) {
				return after.instantAfter();
			}
			else if (standaloneFrom != null) {
				// note: checked last so that it can be overridden by others
				return standaloneFrom;
			}
			return null;
		}

		public PartialDate getTo() {
			if (to != null) {
				return to;
			}
			else if (before != null) {
				return before.instantBefore();
			}
			else if (standaloneTo != null) {
				// note: checked last so that it can be overridden by others
				return standaloneTo;
			}
			return null;
		}
	}
	
	/**
	 * Internal support class to aid in modifying components of dates.
	 * Ensures a number of things in addition to the standard capabilities of 
	 * the {@link Calendar} class:
	 * <ul>
	 * <li> the {@code definedFrom} and {@code definedTo} values of the partial date are not lost,
	 * <li> the inherent component-bias for non-year components is maintained in a way that works
	 *      correctly even for leap-years and leap-seconds.
	 * </ul>
	 * 
	 * <p> Internally stores the date with a 'start' bias plus a collection of post-processors
	 * that re-apply the 'end' bias if applicable. This avoids miscalculations due to differences
	 * across years; eg: the number of days in Februrary between leap/non-leap years. 
	 */
	protected static class PartialDateBuilder implements Comparable<PartialDateBuilder> {
		private Calendar cal;
		private DefinedLevel definedFrom;
		private DefinedLevel definedTo;
		private List<DatePostProcessor> postProcessors = new ArrayList<DatePostProcessor>();
		
		/**
		 * @param date {@code java.lang.Date} or {@code PartialDate} 
		 * @param bias if null then all components are left unchanged and no biasing is applied
		 */
		public PartialDateBuilder(Date date, Bias bias) {
			// initialise date with start-bias
			if (bias == null) {
				this.cal = calendarOf(date);
			}
			else {
				this.cal = applyTo(calendarOf(date), postProcessorsFor(date, Bias.START));
			}
			this.definedFrom = (date instanceof PartialDate) ? ((PartialDate) date).definedFrom() : DefinedLevel.YEAR;
			this.definedTo   = (date instanceof PartialDate) ? ((PartialDate) date).definedTo()   : DefinedLevel.DAY;
			
			// prepare post processor for END bias
			// (don't need if START bias because just did that)
			if (bias != null && bias != Bias.START) {
				postProcessors = postProcessorsFor(date, bias);
			}
		}

		// returns string of final date
		@Override
		public String toString() {
			Calendar finalCal = applyTo(this.cal, postProcessors);
			return finalCal.getTime().toString();
		}
		
		// returns final partial date
		public PartialDate asPartialDate() {
			Calendar finalCal = applyTo(this.cal, postProcessors);
			return new PartialDate(finalCal.getTimeInMillis(), definedFrom, definedTo);
		}

		// return final date
		public Date asDate() {
			return applyTo(this.cal, postProcessors).getTime();
		}
		
		// compares based on final dates
		public int compareTo(PartialDateBuilder other) {
			return asDate().compareTo(other.asDate());
		}
		
		// sets internal non-final date component
		public void set(int field, int value) {
			cal.set(field, value);
		}

		// adds to internal non-final date component
		public void add(int field, int amount) {
			cal.add(field, amount);
		}
		
		// gets internal non-final date component
		public int get(int field) {
			return cal.get(field);
		}
		
		/**
		 * Generates post-processors to apply the given bias to the date, given
		 * how much of the given date is undefined.
		 * No processors are created if not a {@code PartialDate}.
		 */
		private List<DatePostProcessor> postProcessorsFor(Date date, Bias bias) {
			List<DatePostProcessor> processors = new ArrayList<DatePostProcessor>();
			if (date instanceof PartialDate) {
				if (!((PartialDate) date).isDefinedAt(DefinedLevel.MONTH)) {
					processors.add(FlexibleDateFormat.defaultMonthOfYear(bias));
				}
				if (!((PartialDate) date).isDefinedAt(DefinedLevel.DAY)) {
					processors.add(FlexibleDateFormat.defaultDayOfMonth(bias));
				}
				processors.add(FlexibleDateFormat.defaultInstantOfDay(bias));
			}
			return processors;
		}
		
		/**
		 * Applies the given post-processors to the given calendar.
		 */
		private Calendar applyTo(Calendar c, List<DatePostProcessor> processors) {
			if (processors.isEmpty()) {
				return c;
			}
			
			Date date = c.getTime();
			for (DatePostProcessor processor: processors) {
				try {
					date = processor.process(date);
				} catch (ParseException e) {
					// only possible if there's a bug
					throw new UnsupportedOperationException("Unexpected exception ocurred while calculating date: "+e.getMessage(), e);
				}
			}
			return calendarOf(date);
		}
	}
	
	/**
	 * Internal support class to aid in constructing a {@link DateRange} instance
	 * by storing its intended state and making it easy to update
	 * until it's finalised.
	 * 
	 * <p> If constructed via DateRange with {@code java.util.Date} components,
	 * then wraps that as if it's actually using PartialDates.
	 */
	protected static class DateRangeBuilder {
		private PartialDate from;
		private PartialDate to;
		
		public static DateRangeBuilder fromTo(PartialDate from, PartialDate to) {
			DateRangeBuilder range = new DateRangeBuilder();
			range.from = from;
			range.to = to;
			return range;
		}
		
		/** Casts components to PartialDate as needed */
		public static DateRangeBuilder forFormatOf(DateRange dateRange) {
			DateRangeBuilder range = new DateRangeBuilder();
			range.from = partialDateOf(dateRange.getFrom());
			range.to = partialDateOf(dateRange.getTo());
			return range;
		}
		
		@Override
		public String toString() {
			return asDateRange().toString();
		}
		
		public DateRange asDateRange() {
			return DateRange.fromTo(from, to);
		}
		
		public PartialDate getFrom() {
			return from;
		}
		
		public PartialDate getTo() {
			return to;
		}
		
		private static PartialDate partialDateOf(Date date) {
			if (date == null) {
				return null;
			}
			else if (date instanceof PartialDate) {
				return (PartialDate) date;
			}
			else {
				return new PartialDate(date);
			}
		}
		
	}

	/**
	 * Used to update date ranges after initial parsing.
	 * Used to apply additional validation rules and defaulting
	 * mechanisms.
	 */
	protected static interface RangePostProcessor {
		public DateRangeBuilder process(DateRangeBuilder range) throws ParseException;
	}
}
