/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.util.Calendar;
import java.util.Date;

import org.apache.commons.lang3.time.DateUtils;

/**
 * Represents a partially defined date.
 * Dates are defined <i>from</i> their most significant component
 * <i>to</i> their least significant component.
 * The date will have values in more significant and in less significant
 * components than it is defined, however those values will have been
 * supplied by defaulting mechanisms.
 * 
 * <p> The {@link #equals(Object)} and {@link #hashCode()} methods
 * return as per the standard {@link Date} type.
 * 
 * <p> Note: At present it is assumed that dates have no time components
 * but this is subject to change in a later version.  
 * @author Malcolm Lett
 */
public class PartialDate extends Date {
	private static final long serialVersionUID = 1L;
	private static final DefinedLevel BROADEST_DEFINITION = DefinedLevel.YEAR;
	private static final DefinedLevel FINEST_DEFINITION = DefinedLevel.DAY;
	private DefinedLevel definedFrom;
	private DefinedLevel definedTo;
	
	/**
	 * New fully defined instance.
	 * @param value the date value
	 */
	public PartialDate(Date value) {
		this(value,
				(value instanceof PartialDate) ? ((PartialDate) value).definedFrom() : null,
				(value instanceof PartialDate) ? ((PartialDate) value).definedTo() : null);
	}

	/**
	 * New fully defined instance.
	 * @param value the date value
	 */
	public PartialDate(long value) {
		this(value, null, null);
	}
	
	/**
	 * New instance with specified definition level.
	 * @param value the date value
	 * @param definedFrom level from which this date is defined
	 * @param definedTo level to which this date is defined
	 */
	public PartialDate(Date value, DefinedLevel definedFrom, DefinedLevel definedTo) {
		this(value.getTime(), definedFrom, definedTo);
	}

	/**
	 * New instance with specified definition level.
	 * @param time the date value in milliseconds since the epoc
	 * @param definedFrom level from which this date is defined
	 * @param definedTo level to which this date is defined
	 */
	public PartialDate(long time, DefinedLevel definedFrom, DefinedLevel definedTo) {
		super(time);
		this.definedFrom = (definedFrom == null) ? BROADEST_DEFINITION : definedFrom;
		this.definedTo = (definedTo == null) ? FINEST_DEFINITION : definedTo;
	}
	
	/**
	 * Converts this instance to a standard {@code Date} object.
	 * @return values as a {@link java.util.Date}
	 */
	public Date asDate() {
		return new Date(this.getTime());
	}

	/**
	 * Indicates the level from which the date was originally defined
	 * when specified. The date will have values in more significant
	 * components, however they will have been from defaulting mechanisms.
	 * @return the most significant component which has been defined
	 */
	public DefinedLevel definedFrom() {
		return definedFrom;
	}
	
	/**
	 * Indicates the level to which the date was originally defined
	 * when specified. The date may have values in less significant
	 * components, however they will have been from defaulting mechanisms.
	 * @return the least significant component which has been defined
	 */
	public DefinedLevel definedTo() {
		return definedTo;
	}
	
	/**
	 * Indicates whether the given level is defined.
	 * This is a convenience method that uses the {@link #definedFrom()}
	 * and {@link #definedTo()} range.
	 * @param level the level to check
	 */
	public boolean isDefinedAt(DefinedLevel level) {
		return definedFrom().isMoreSignificantThanOrEqualTo(level) && definedTo().isLessSignificantThanOrEqualTo(level);
	}
	
	/**
	 * Gets a new instance with the same value as this date but
	 * with the specified level.
	 * @param level new level from which the date should be defined
	 * @return date with new defined levels
	 */
	public PartialDate defineFrom(DefinedLevel level) {
		if (definedFrom == level) {
			return this;
		}
		return new PartialDate(getTime(), level, definedTo);
	}
	
	/**
	 * Gets a new instance with the same value as this date but
	 * with the specified level.
	 * @param level new level to which the date should be defined
	 * @return date with new defined levels
	 */
	public PartialDate defineTo(DefinedLevel level) {
		if (definedTo == level) {
			return this;
		}
		return new PartialDate(getTime(), definedFrom, level);
	}

	/**
	 * Gets a cloned instance with the date set to the smallest measurable unit
	 * before this date.
	 * <p> The returned partial date is defined from the same level to the same level
	 * as this one.
	 * @return new date instant at the instant before this one
	 */
	public PartialDate instantBefore() {
		return plusMillis(-1);
	}

	/**
	 * Gets a cloned instance with the date set to the smallest measurable unit
	 * after this date.
	 * <p> The returned partial date is defined from the same level to the same level
	 * as this one.
	 * @return new date instant at the instant after this one
	 */
	public PartialDate instantAfter() {
		return plusMillis(+1);
	}
	
	/**
	 * Gets a cloned instance with the date incremented by the specified number of milliseconds.
	 * Use negative to decrement the date.
	 * <p> The returned partial date is defined from the same level to the same level
	 * as this one.
	 * @param ms number of ms to add, negative to go backwards
	 */
	public PartialDate plusMillis(int ms) {
		return new PartialDate(DateUtils.addMilliseconds(this, ms), definedFrom, definedTo);
	}
	
	/**
	 * Gets a cloned instance with the date incremented by the specified number of days.
	 * Use negative to decrement the date.
	 * <p> The returned partial date is defined from the same level to the same level
	 * as this one.
	 * @param days number of days to add, negative to go backwards
	 */
	public PartialDate plusDays(int days) {
		return new PartialDate(DateUtils.addDays(this, days), definedFrom, definedTo);
	}

	/**
	 * Gets a cloned instance with the date incremented by the specified number of months.
	 * Use negative to decrement the date.
	 * <p> The returned partial date is defined from the same level to the same level
	 * as this one.
	 * @param months number of months to add, negative to go backwards
	 */
	public PartialDate plusMonths(int months) {
		return new PartialDate(DateUtils.addMonths(this, months), definedFrom, definedTo);
	}

	/**
	 * Gets a cloned instance with the date incremented by the specified number of years.
	 * Use negative to decrement the date.
	 * <p> The returned partial date is defined from the same level to the same level
	 * as this one.
	 * @param years number of years to add, negative to go backwards
	 */
	public PartialDate plusYears(int years) {
		return new PartialDate(DateUtils.addYears(this, years), definedFrom, definedTo);
	}

	/**
	 * Compares the two dates.
	 * All parts of the date are considered, regardless of which components are defined.
	 * @param other the other date to compare to
	 */
	public int compareTo(PartialDate other) {
		return super.compareTo(other);
	}
	
	/**
	 * Compares the two dates to the specified level.
	 * For example, if {@code toLevel=year} then just requires that the two dates have
	 * the same year in order for them to be considered equal.
	 * @param other the other date to compare to
	 * @param toLevel level to which the comparison should be made (from year down to the defined level)
	 */
	public int compareTo(PartialDate other, DefinedLevel toLevel) {
		Calendar c1 = DateUtils.toCalendar(this);
		Calendar c2 = DateUtils.toCalendar(other);
		int result = c1.get(Calendar.YEAR) - c2.get(Calendar.YEAR);
		if (result != 0) {
			return result;
		}
		
		if (toLevel.isLessSignificantThanOrEqualTo(DefinedLevel.MONTH)) {
			result = c1.get(Calendar.MONTH) - c2.get(Calendar.MONTH);
			if (result != 0) {
				return result;
			}
		}

		if (toLevel.isLessSignificantThanOrEqualTo(DefinedLevel.DAY)) {
			result = c1.get(Calendar.DAY_OF_MONTH) - c2.get(Calendar.DAY_OF_MONTH);
			if (result != 0) {
				return result;
			}
		}

		if (toLevel.isLessSignificantThan(DefinedLevel.DAY)) {
			result = c1.get(Calendar.HOUR_OF_DAY) - c2.get(Calendar.HOUR_OF_DAY);
			if (result != 0) {
				return result;
			}

			result = c1.get(Calendar.MINUTE) - c2.get(Calendar.MINUTE);
			if (result != 0) {
				return result;
			}

			result = c1.get(Calendar.SECOND) - c2.get(Calendar.SECOND);
			if (result != 0) {
				return result;
			}
		
			result = c1.get(Calendar.MILLISECOND) - c2.get(Calendar.MILLISECOND);
			if (result != 0) {
				return result;
			}
		}
		return 0;
	}
}

