/*
 * Copyright 2013 Malcolm Lett
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.tedhi;

import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * Support class to make it easy to peek around lists.
 * @param <T> type of collection
 */
class PeekableListIterator<T> implements ListIterator<T> {
	private ListIterator<T> delegate;
	//private List<T> peekCache = new LinkedList<T>();

	public static <E> PeekableListIterator<E> of(List<E> list) {
		return new PeekableListIterator<E>(list.listIterator());
	}

	public static <E> PeekableListIterator<E> of(ListIterator<E> listIterator) {
		return new PeekableListIterator<E>(listIterator);
	}
	
	private PeekableListIterator(ListIterator<T> delegate) {
		this.delegate = delegate;
	}

	/**
	 * @return the next element without consuming it
	 * @throws NoSuchElementException if the iteration has no next element
	 */
	public T peekNext() {
		T item = delegate.next();
		
		// move back
		delegate.previous();
		return item;
	}

	/**
	 * @return the previous element without consuming it
	 * @throws NoSuchElementException if the iteration has no previous element
	 */
	public T peekPrevious() {
		T item = delegate.previous();
		
		// move back
		delegate.next();
		return item;
	}
	
	/**
	 * Note: 0 is considered to refer to the current one, which is not accessible.
	 * @param numAhead
	 * @return the next n-th element without consuming it
	 * @throws NoSuchElementException if the iteration has no such next element
	 */
	public T peekNext(int numAhead) {
		if (numAhead == 0) {
			throw new IllegalArgumentException();
		}
		if (numAhead < 0) {
			return peekPrevious(-numAhead);
		}
		
		T item = null;
		for (int cnt = 0; cnt < numAhead; cnt++) {
			item = delegate.next();
		}
		
		// move back
		for (int cnt = 0; cnt < numAhead; cnt++) {
			delegate.previous();
		}
		return item;
	}

	/**
	 * Note: 0 is considered to refer to the current one, which is not accessible.
	 * @param numBehind
	 * @return the next n-th element without consuming it
	 * @throws NoSuchElementException if the iteration has no such previous element
	 */
	public T peekPrevious(int numBehind) {
		if (numBehind == 0) {
			throw new IllegalArgumentException();
		}
		if (numBehind < 0) {
			return peekNext(-numBehind);
		}
		
		T item = null;
		for (int cnt = 0; cnt < numBehind; cnt++) {
			item = delegate.previous();
		}
		
		// move back
		for (int cnt = 0; cnt < numBehind; cnt++) {
			delegate.next();
		}
		return item;
	}
	
	/**
	 * Note: 0 is treated as the current which cannot be accessed.
	 * @param offset
	 * @return
	 */
	public boolean hasNext(int offset) {
		if (offset == 0) {
			return false;
		}
		if (offset < 0) {
			return hasPrevious(-offset);
		}

		int countMoved = 0;
		for (int cnt = 0; cnt < offset && delegate.hasNext(); cnt++) {
			delegate.next();
			countMoved++;
		}
		
		// move back
		for (int cnt = 0; cnt < countMoved; cnt++) {
			delegate.previous();
		}
		
		return (countMoved >= offset);
	}

	/**
	 * Note: 0 is treated as the current which cannot be accessed.
	 * @param offset
	 * @return
	 */
	public boolean hasPrevious(int offset) {
		if (offset == 0) {
			return false;
		}
		if (offset < 0) {
			return hasNext(-offset);
		}

		int countMoved = 0;
		for (int cnt = 0; cnt < offset && delegate.hasPrevious(); cnt++) {
			delegate.previous();
			countMoved++;
		}
		
		// move back
		for (int cnt = 0; cnt < countMoved; cnt++) {
			delegate.next();
		}
		
		return (countMoved >= offset);
	}
	
	public boolean hasNext() {
		return delegate.hasNext();
	}

	public T next() {
		return delegate.next();
	}

	public boolean hasPrevious() {
		return delegate.hasPrevious();
	}

	public T previous() {
		return delegate.previous();
	}

	public int nextIndex() {
		return delegate.nextIndex();
	}

	public int previousIndex() {
		return delegate.previousIndex();
	}

	public void remove() {
		delegate.remove();
	}

	public void set(T e) {
		delegate.set(e);
	}

	public void add(T e) {
		delegate.add(e);
	}

}
